"""Tests for geometry module: Vector3D, Polygon3D, and utility functions."""

from __future__ import annotations

import pytest

from idfkit import IDFDocument
from idfkit.geometry import (
    Polygon3D,
    Vector3D,
    _is_convex_2d,
    _point_in_polygon_2d,
    calculate_surface_area,
    calculate_zone_floor_area,
    get_surface_coords,
    get_zone_origin,
    get_zone_rotation,
    polygon_area_2d,
    polygon_contains_2d,
    polygon_difference_2d,
    polygon_intersection_2d,
    set_surface_coords,
)
from idfkit.objects import IDFObject

_TOL = 1e-7


def _close(a: float, b: float, tol: float = _TOL) -> bool:
    return abs(a - b) < tol


# ---------------------------------------------------------------------------
# Vector3D
# ---------------------------------------------------------------------------


class TestVector3D:
    def test_create(self) -> None:
        v = Vector3D(1.0, 2.0, 3.0)
        assert v.x == 1.0
        assert v.y == 2.0
        assert v.z == 3.0

    def test_immutable(self) -> None:
        v = Vector3D(1.0, 2.0, 3.0)
        with pytest.raises(AttributeError):
            v.x = 5.0  # type: ignore[misc]

    def test_add(self) -> None:
        v1 = Vector3D(1, 2, 3)
        v2 = Vector3D(4, 5, 6)
        result = v1 + v2
        assert result == Vector3D(5, 7, 9)

    def test_sub(self) -> None:
        v1 = Vector3D(4, 5, 6)
        v2 = Vector3D(1, 2, 3)
        result = v1 - v2
        assert result == Vector3D(3, 3, 3)

    def test_mul(self) -> None:
        v = Vector3D(1, 2, 3)
        result = v * 2
        assert result == Vector3D(2, 4, 6)

    def test_rmul(self) -> None:
        v = Vector3D(1, 2, 3)
        result = 3 * v
        assert result == Vector3D(3, 6, 9)

    def test_truediv(self) -> None:
        v = Vector3D(6, 8, 10)
        result = v / 2
        assert result == Vector3D(3, 4, 5)

    def test_neg(self) -> None:
        v = Vector3D(1, -2, 3)
        result = -v
        assert result == Vector3D(-1, 2, -3)

    def test_dot(self) -> None:
        v1 = Vector3D(1, 0, 0)
        v2 = Vector3D(0, 1, 0)
        assert v1.dot(v2) == 0.0  # orthogonal

    def test_dot_parallel(self) -> None:
        v1 = Vector3D(1, 2, 3)
        assert v1.dot(v1) == 14.0

    def test_cross(self) -> None:
        v1 = Vector3D(1, 0, 0)
        v2 = Vector3D(0, 1, 0)
        result = v1.cross(v2)
        assert result == Vector3D(0, 0, 1)

    def test_cross_anticommutative(self) -> None:
        v1 = Vector3D(1, 0, 0)
        v2 = Vector3D(0, 1, 0)
        assert v1.cross(v2) == -(v2.cross(v1))

    def test_length(self) -> None:
        v = Vector3D(3, 4, 0)
        assert _close(v.length(), 5.0)

    def test_length_unit(self) -> None:
        v = Vector3D(1, 0, 0)
        assert _close(v.length(), 1.0)

    def test_normalize(self) -> None:
        v = Vector3D(3, 4, 0)
        n = v.normalize()
        assert _close(n.length(), 1.0)
        assert _close(n.x, 0.6)
        assert _close(n.y, 0.8)

    def test_normalize_zero_vector(self) -> None:
        v = Vector3D(0, 0, 0)
        n = v.normalize()
        assert n == Vector3D(0, 0, 0)

    def test_rotate_z_90(self) -> None:
        v = Vector3D(1, 0, 0)
        rotated = v.rotate_z(90)
        assert _close(rotated.x, 0.0, 1e-10)
        assert _close(rotated.y, 1.0)
        assert _close(rotated.z, 0.0)

    def test_rotate_z_180(self) -> None:
        v = Vector3D(1, 0, 0)
        rotated = v.rotate_z(180)
        assert _close(rotated.x, -1.0)
        assert _close(rotated.y, 0.0, 1e-10)

    def test_rotate_z_preserves_z(self) -> None:
        v = Vector3D(1, 0, 5)
        rotated = v.rotate_z(45)
        assert _close(rotated.z, 5.0)

    def test_as_tuple(self) -> None:
        v = Vector3D(1, 2, 3)
        assert v.as_tuple() == (1, 2, 3)

    def test_from_tuple(self) -> None:
        v = Vector3D.from_tuple((1, 2, 3))
        assert v == Vector3D(1, 2, 3)

    def test_from_tuple_list(self) -> None:
        v = Vector3D.from_tuple([1.0, 2.0, 3.0])
        assert v == Vector3D(1, 2, 3)

    def test_origin(self) -> None:
        v = Vector3D.origin()
        assert v == Vector3D(0, 0, 0)


# ---------------------------------------------------------------------------
# Polygon3D
# ---------------------------------------------------------------------------


class TestPolygon3D:
    @pytest.fixture
    def unit_square(self) -> Polygon3D:
        """A 1x1 horizontal square at z=0."""
        return Polygon3D([
            Vector3D(0, 0, 0),
            Vector3D(1, 0, 0),
            Vector3D(1, 1, 0),
            Vector3D(0, 1, 0),
        ])

    @pytest.fixture
    def vertical_wall(self) -> Polygon3D:
        """A 10x3 vertical wall in the XZ plane."""
        return Polygon3D([
            Vector3D(0, 0, 3),
            Vector3D(0, 0, 0),
            Vector3D(10, 0, 0),
            Vector3D(10, 0, 3),
        ])

    def test_num_vertices(self, unit_square: Polygon3D) -> None:
        assert unit_square.num_vertices == 4

    def test_area_unit_square(self, unit_square: Polygon3D) -> None:
        assert _close(unit_square.area, 1.0)

    def test_area_vertical_wall(self, vertical_wall: Polygon3D) -> None:
        assert _close(vertical_wall.area, 30.0)

    def test_area_degenerate(self) -> None:
        poly = Polygon3D([Vector3D(0, 0, 0), Vector3D(1, 0, 0)])
        assert poly.area == 0.0

    def test_area_triangle(self) -> None:
        poly = Polygon3D([
            Vector3D(0, 0, 0),
            Vector3D(4, 0, 0),
            Vector3D(0, 3, 0),
        ])
        assert _close(poly.area, 6.0)

    def test_normal_horizontal(self, unit_square: Polygon3D) -> None:
        n = unit_square.normal
        assert abs(n.z) > 0.99  # should be pointing up or down

    def test_normal_vertical(self, vertical_wall: Polygon3D) -> None:
        n = vertical_wall.normal
        assert abs(n.z) < 0.01  # should be horizontal

    def test_normal_degenerate(self) -> None:
        poly = Polygon3D([Vector3D(0, 0, 0)])
        n = poly.normal
        assert n == Vector3D(0, 0, 1)  # default for degenerate

    def test_centroid(self, unit_square: Polygon3D) -> None:
        c = unit_square.centroid
        assert _close(c.x, 0.5)
        assert _close(c.y, 0.5)
        assert _close(c.z, 0.0)

    def test_centroid_empty(self) -> None:
        poly = Polygon3D([])
        c = poly.centroid
        assert c == Vector3D.origin()

    def test_is_horizontal(self, unit_square: Polygon3D) -> None:
        assert unit_square.is_horizontal is True

    def test_is_not_horizontal(self, vertical_wall: Polygon3D) -> None:
        assert vertical_wall.is_horizontal is False

    def test_is_vertical(self, vertical_wall: Polygon3D) -> None:
        assert vertical_wall.is_vertical is True

    def test_is_not_vertical(self, unit_square: Polygon3D) -> None:
        assert unit_square.is_vertical is False

    def test_translate(self, unit_square: Polygon3D) -> None:
        offset = Vector3D(5, 5, 5)
        translated = unit_square.translate(offset)
        assert translated.vertices[0] == Vector3D(5, 5, 5)
        assert translated.vertices[2] == Vector3D(6, 6, 5)

    def test_rotate_z(self, unit_square: Polygon3D) -> None:
        rotated = unit_square.rotate_z(90)
        # After 90-degree rotation around centroid, area should be preserved
        assert _close(rotated.area, unit_square.area)

    def test_rotate_z_with_anchor(self) -> None:
        poly = Polygon3D([
            Vector3D(1, 0, 0),
            Vector3D(2, 0, 0),
            Vector3D(2, 1, 0),
            Vector3D(1, 1, 0),
        ])
        anchor = Vector3D(0, 0, 0)
        rotated = poly.rotate_z(90, anchor=anchor)
        # After rotating 90 degrees around origin, (1,0) -> (0,1)
        assert _close(rotated.vertices[0].x, 0.0, 1e-10)
        assert _close(rotated.vertices[0].y, 1.0)

    def test_as_tuple_list(self, unit_square: Polygon3D) -> None:
        tuples = unit_square.as_tuple_list()
        assert tuples == [(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0)]

    def test_from_tuples(self) -> None:
        coords = [(0, 0, 0), (1, 0, 0), (1, 1, 0)]
        poly = Polygon3D.from_tuples(coords)
        assert poly.num_vertices == 3
        assert poly.vertices[0] == Vector3D(0, 0, 0)


# ---------------------------------------------------------------------------
# Surface geometry utilities
# ---------------------------------------------------------------------------


class TestSurfaceGeometryUtils:
    def test_get_surface_coords(self) -> None:
        surface = IDFObject(
            obj_type="BuildingSurface:Detailed",
            name="Wall",
            data={
                "number_of_vertices": 4,
                "vertex_1_x_coordinate": 0.0,
                "vertex_1_y_coordinate": 0.0,
                "vertex_1_z_coordinate": 3.0,
                "vertex_2_x_coordinate": 0.0,
                "vertex_2_y_coordinate": 0.0,
                "vertex_2_z_coordinate": 0.0,
                "vertex_3_x_coordinate": 10.0,
                "vertex_3_y_coordinate": 0.0,
                "vertex_3_z_coordinate": 0.0,
                "vertex_4_x_coordinate": 10.0,
                "vertex_4_y_coordinate": 0.0,
                "vertex_4_z_coordinate": 3.0,
            },
        )
        poly = get_surface_coords(surface)
        assert poly is not None
        assert poly.num_vertices == 4
        assert _close(poly.area, 30.0)

    def test_get_surface_coords_autodetect_vertices(self) -> None:
        """Test that vertices are autodetected when number_of_vertices is missing."""
        surface = IDFObject(
            obj_type="BuildingSurface:Detailed",
            name="Wall",
            data={
                "vertex_1_x_coordinate": 0.0,
                "vertex_1_y_coordinate": 0.0,
                "vertex_1_z_coordinate": 0.0,
                "vertex_2_x_coordinate": 1.0,
                "vertex_2_y_coordinate": 0.0,
                "vertex_2_z_coordinate": 0.0,
                "vertex_3_x_coordinate": 1.0,
                "vertex_3_y_coordinate": 1.0,
                "vertex_3_z_coordinate": 0.0,
            },
        )
        poly = get_surface_coords(surface)
        assert poly is not None
        assert poly.num_vertices == 3

    def test_get_surface_coords_blank_number_of_vertices(self) -> None:
        """Test that blank number_of_vertices is treated as autocalculate."""
        surface = IDFObject(
            obj_type="BuildingSurface:Detailed",
            name="Wall",
            data={
                "number_of_vertices": "",  # blank = autocalculate in EnergyPlus
                "vertex_1_x_coordinate": 0.0,
                "vertex_1_y_coordinate": 0.0,
                "vertex_1_z_coordinate": 3.0,
                "vertex_2_x_coordinate": 0.0,
                "vertex_2_y_coordinate": 0.0,
                "vertex_2_z_coordinate": 0.0,
                "vertex_3_x_coordinate": 10.0,
                "vertex_3_y_coordinate": 0.0,
                "vertex_3_z_coordinate": 0.0,
                "vertex_4_x_coordinate": 10.0,
                "vertex_4_y_coordinate": 0.0,
                "vertex_4_z_coordinate": 3.0,
            },
        )
        poly = get_surface_coords(surface)
        assert poly is not None
        assert poly.num_vertices == 4
        assert _close(poly.area, 30.0)

    def test_get_surface_coords_no_vertices(self) -> None:
        surface = IDFObject(obj_type="BuildingSurface:Detailed", name="Empty", data={})
        poly = get_surface_coords(surface)
        assert poly is None

    def test_get_surface_coords_schema_naming(self) -> None:
        """Test epJSON schema naming: vertex_x_coordinate, vertex_x_coordinate_2, ..."""
        surface = IDFObject(
            obj_type="BuildingSurface:Detailed",
            name="SchemaWall",
            data={
                "number_of_vertices": 4,
                "vertex_x_coordinate": 0.0,
                "vertex_y_coordinate": 0.0,
                "vertex_z_coordinate": 3.0,
                "vertex_x_coordinate_2": 0.0,
                "vertex_y_coordinate_2": 0.0,
                "vertex_z_coordinate_2": 0.0,
                "vertex_x_coordinate_3": 10.0,
                "vertex_y_coordinate_3": 0.0,
                "vertex_z_coordinate_3": 0.0,
                "vertex_x_coordinate_4": 10.0,
                "vertex_y_coordinate_4": 0.0,
                "vertex_z_coordinate_4": 3.0,
            },
        )
        poly = get_surface_coords(surface)
        assert poly is not None
        assert poly.num_vertices == 4
        assert _close(poly.area, 30.0)

    def test_set_surface_coords(self) -> None:
        surface = IDFObject(obj_type="BuildingSurface:Detailed", name="Wall", data={})
        poly = Polygon3D([
            Vector3D(0, 0, 0),
            Vector3D(1, 0, 0),
            Vector3D(1, 1, 0),
            Vector3D(0, 1, 0),
        ])
        set_surface_coords(surface, poly)
        assert surface.number_of_vertices == 4
        assert surface.vertex_1_x_coordinate == 0.0
        assert surface.vertex_4_y_coordinate == 1.0

    def test_get_zone_origin(self) -> None:
        zone = IDFObject(
            obj_type="Zone",
            name="Z",
            data={"x_origin": 10.0, "y_origin": 20.0, "z_origin": 5.0},
        )
        origin = get_zone_origin(zone)
        assert origin == Vector3D(10, 20, 5)

    def test_get_zone_origin_defaults(self) -> None:
        zone = IDFObject(obj_type="Zone", name="Z", data={})
        origin = get_zone_origin(zone)
        assert origin == Vector3D(0, 0, 0)

    def test_get_zone_rotation(self) -> None:
        zone = IDFObject(obj_type="Zone", name="Z", data={"direction_of_relative_north": 45.0})
        assert get_zone_rotation(zone) == 45.0

    def test_get_zone_rotation_default(self) -> None:
        zone = IDFObject(obj_type="Zone", name="Z", data={})
        assert get_zone_rotation(zone) == 0.0

    def test_calculate_surface_area(self, simple_doc: IDFDocument) -> None:
        wall = simple_doc.getobject("BuildingSurface:Detailed", "TestWall")
        assert wall is not None
        area = calculate_surface_area(wall)
        assert _close(area, 30.0)

    def test_calculate_surface_area_no_coords(self) -> None:
        surface = IDFObject(obj_type="BuildingSurface:Detailed", name="Empty", data={})
        assert calculate_surface_area(surface) == 0.0

    def test_calculate_zone_floor_area(self, simple_doc: IDFDocument) -> None:
        area = calculate_zone_floor_area(simple_doc, "TestZone")
        assert _close(area, 100.0)

    def test_calculate_zone_floor_area_no_floors(self, empty_doc: IDFDocument) -> None:
        empty_doc.add("Zone", "EmptyZone")
        area = calculate_zone_floor_area(empty_doc, "EmptyZone")
        assert area == 0.0


# ---------------------------------------------------------------------------
# 2D polygon operations
# ---------------------------------------------------------------------------


class TestPolygonIntersection2D:
    def test_overlapping_rectangles(self) -> None:
        # Two overlapping rectangles
        a = [(0, 0), (10, 0), (10, 10), (0, 10)]
        b = [(5, 5), (15, 5), (15, 15), (5, 15)]
        result = polygon_intersection_2d(a, b)
        assert result is not None
        area = abs(polygon_area_2d(result))
        assert abs(area - 25.0) < 0.01  # 5x5 overlap

    def test_contained_rectangle(self) -> None:
        # Inner fully inside outer
        outer = [(0, 0), (20, 0), (20, 20), (0, 20)]
        inner = [(5, 5), (15, 5), (15, 15), (5, 15)]
        result = polygon_intersection_2d(outer, inner)
        assert result is not None
        area = abs(polygon_area_2d(result))
        assert abs(area - 100.0) < 0.01  # inner is 10x10

    def test_disjoint_returns_none(self) -> None:
        a = [(0, 0), (5, 0), (5, 5), (0, 5)]
        b = [(10, 10), (15, 10), (15, 15), (10, 15)]
        result = polygon_intersection_2d(a, b)
        assert result is None

    def test_concave_subject_convex_clip(self) -> None:
        # L-shape clipped by rectangle
        from idfkit.zoning import footprint_l_shape

        l_shape = footprint_l_shape(20, 10, 8, 5)
        clip = [(0, 0), (10, 0), (10, 15), (0, 15)]
        result = polygon_intersection_2d(l_shape, clip)
        assert result is not None
        assert abs(polygon_area_2d(result)) > 0


class TestPolygonDifference2D:
    def test_rectangles_frame(self) -> None:
        outer = [(0, 0), (20, 0), (20, 20), (0, 20)]
        inner = [(5, 5), (15, 5), (15, 15), (5, 15)]
        result = polygon_difference_2d(outer, inner)
        assert result is not None
        # The result is a slit/bridge polygon containing vertices from both
        # outer and inner polygons.
        assert len(result) == len(outer) + len(inner)
        # The conceptual frame area can be computed from the original polygons.
        frame_area = abs(polygon_area_2d(outer)) - abs(polygon_area_2d(inner))
        assert abs(frame_area - 300.0) < 0.01

    def test_same_polygon_returns_none(self) -> None:
        poly = [(0, 0), (10, 0), (10, 10), (0, 10)]
        result = polygon_difference_2d(poly, poly)
        assert result is None


class TestPolygonContains2D:
    def test_contained(self) -> None:
        outer = [(0, 0), (20, 0), (20, 20), (0, 20)]
        inner = [(5, 5), (15, 5), (15, 15), (5, 15)]
        assert polygon_contains_2d(outer, inner)

    def test_not_contained(self) -> None:
        outer = [(0, 0), (10, 0), (10, 10), (0, 10)]
        inner = [(5, 5), (15, 5), (15, 15), (5, 15)]
        assert not polygon_contains_2d(outer, inner)


class TestPolygonArea2D:
    def test_unit_square(self) -> None:
        poly = [(0, 0), (1, 0), (1, 1), (0, 1)]
        area = polygon_area_2d(poly)
        assert abs(abs(area) - 1.0) < 1e-10

    def test_rectangle(self) -> None:
        poly = [(0, 0), (5, 0), (5, 3), (0, 3)]
        area = polygon_area_2d(poly)
        assert abs(abs(area) - 15.0) < 1e-10

    def test_triangle(self) -> None:
        poly = [(0, 0), (4, 0), (0, 3)]
        area = polygon_area_2d(poly)
        assert abs(abs(area) - 6.0) < 1e-10

    def test_signed_area_ccw_positive(self) -> None:
        # Counter-clockwise winding should give positive signed area
        ccw = [(0, 0), (1, 0), (1, 1), (0, 1)]
        assert polygon_area_2d(ccw) > 0
        # Clockwise winding should give negative signed area
        cw = [(0, 0), (0, 1), (1, 1), (1, 0)]
        assert polygon_area_2d(cw) < 0


class TestIsConvex2D:
    def test_square_is_convex(self) -> None:
        poly = [(0, 0), (1, 0), (1, 1), (0, 1)]
        assert _is_convex_2d(poly)

    def test_l_shape_not_convex(self) -> None:
        poly = [(0, 0), (2, 0), (2, 1), (1, 1), (1, 2), (0, 2)]
        assert not _is_convex_2d(poly)

    def test_triangle_is_convex(self) -> None:
        poly = [(0, 0), (4, 0), (2, 3)]
        assert _is_convex_2d(poly)


class TestPointInPolygon2D:
    def test_inside(self) -> None:
        poly = [(0, 0), (10, 0), (10, 10), (0, 10)]
        assert _point_in_polygon_2d((5, 5), poly)

    def test_outside(self) -> None:
        poly = [(0, 0), (10, 0), (10, 10), (0, 10)]
        assert not _point_in_polygon_2d((15, 15), poly)

    def test_on_edge(self) -> None:
        poly = [(0, 0), (10, 0), (10, 10), (0, 10)]
        # Points on edges may return True or False depending on implementation;
        # just ensure no crash
        _point_in_polygon_2d((5, 0), poly)

    def test_inside_triangle(self) -> None:
        poly = [(0, 0), (10, 0), (5, 10)]
        assert _point_in_polygon_2d((5, 3), poly)

    def test_outside_triangle(self) -> None:
        poly = [(0, 0), (10, 0), (5, 10)]
        assert not _point_in_polygon_2d((0, 10), poly)



from __future__ import annotations

from idfkit.simulation import TimeSeriesResult
from typing import Any

fig: Any = ...  # type: ignore[assignment]
ts: TimeSeriesResult = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.simulation import MatplotlibBackend

backend = MatplotlibBackend()
fig = backend.line(
    x=list(ts.timestamps),
    y=list(ts.values),
    title="Zone Temperature",
    xlabel="Time",
    ylabel="Temperature (°C)",
)

# Save to file
fig.savefig("temperature.png")
# --8<-- [end:example]



"""Tests for the expand_objects, run_slab_preprocessor, and run_basement_preprocessor functionality."""

from __future__ import annotations

import shutil
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from idfkit import IDFDocument, new_document
from idfkit.exceptions import ExpandObjectsError
from idfkit.objects import IDFObject
from idfkit.simulation.config import EnergyPlusConfig
from idfkit.simulation.expand import (
    _check_for_fatal_preprocessor_message,
    _check_output_not_empty,
    _check_process_exit_code,
    _has_basement_objects,
    _has_slab_objects,
    _needs_expansion,
    expand_objects,
    needs_ground_heat_preprocessing,
    run_basement_preprocessor,
    run_preprocessing,
    run_slab_preprocessor,
)

# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def model_with_hvac_template() -> IDFDocument:
    """Create a document with an HVACTemplate object."""
    doc = new_document(version=(24, 1, 0))
    doc.add("Zone", "Office", {"x_origin": 0.0, "y_origin": 0.0, "z_origin": 0.0})
    doc.add(
        "HVACTemplate:Zone:IdealLoadsAirSystem",
        "Office Ideal Loads",
        {"zone_name": "Office"},
        validate=False,
    )
    return doc


@pytest.fixture
def mock_config(tmp_path: Path) -> EnergyPlusConfig:
    """Create a mock EnergyPlusConfig with fake preprocessor executables."""
    exe = tmp_path / "energyplus"
    exe.touch()
    idd = tmp_path / "Energy+.idd"
    idd.write_text("!IDD_Version 24.1.0\n")

    # ExpandObjects
    expand_exe = tmp_path / "ExpandObjects"
    expand_exe.touch()
    expand_exe.chmod(0o755)

    # Slab & Basement
    preprocess = tmp_path / "PreProcess" / "GrndTempCalc"
    preprocess.mkdir(parents=True)
    slab_exe = preprocess / "Slab"
    slab_exe.touch()
    slab_exe.chmod(0o755)
    (preprocess / "SlabGHT.idd").write_text("! Slab IDD\n")
    basement_exe = preprocess / "Basement"
    basement_exe.touch()
    basement_exe.chmod(0o755)
    (preprocess / "BasementGHT.idd").write_text("! Basement IDD\n")

    return EnergyPlusConfig(
        executable=exe,
        version=(24, 1, 0),
        install_dir=tmp_path,
        idd_path=idd,
    )


def _fake_expand_run(expanded_content: str = "Version, 24.1;\n\nZone,\n  Office;\n"):
    """Return a fake subprocess.run side-effect that writes expanded.idf."""

    def side_effect(cmd: list[str], **kwargs: object) -> MagicMock:
        cwd = Path(str(kwargs.get("cwd", "")))
        (cwd / "expanded.idf").write_text(expanded_content)
        result = MagicMock()
        result.returncode = 0
        result.stdout = ""
        result.stderr = ""
        return result

    return side_effect


# ---------------------------------------------------------------------------
# expand_objects tests
# ---------------------------------------------------------------------------


class TestExpandObjects:
    def test_calls_executable(
        self,
        model_with_hvac_template: IDFDocument,
        mock_config: EnergyPlusConfig,
    ) -> None:
        expanded_idf_content = """\
Version, 24.1;

Zone,
  Office,
  0,
  0, 0, 0,
  1,
  1;

ZoneHVAC:IdealLoadsAirSystem,
  Office Ideal Loads Air,
  ,
  ,
  ,
  50,
  ,
  13,
  ,
  ,
  ,
  ,
  ,
  ;
"""
        with patch("idfkit.simulation.expand.subprocess.run", side_effect=_fake_expand_run(expanded_idf_content)):
            expanded = expand_objects(model_with_hvac_template, energyplus=mock_config)

        assert "ZoneHVAC:IdealLoadsAirSystem" in expanded
        assert expanded["ZoneHVAC:IdealLoadsAirSystem"].first().name == "Office Ideal Loads Air"

    def test_no_expand_exe_raises(self, model_with_hvac_template: IDFDocument, tmp_path: Path) -> None:
        exe = tmp_path / "energyplus"
        exe.touch()
        idd = tmp_path / "Energy+.idd"
        idd.write_text("!IDD_Version 24.1.0\n")
        config = EnergyPlusConfig(executable=exe, version=(24, 1, 0), install_dir=tmp_path, idd_path=idd)

        with pytest.raises(ExpandObjectsError, match="ExpandObjects executable not found") as exc_info:
            expand_objects(model_with_hvac_template, energyplus=config)
        assert exc_info.value.preprocessor == "ExpandObjects"

    def test_no_expanded_file_raises(
        self, model_with_hvac_template: IDFDocument, mock_config: EnergyPlusConfig
    ) -> None:
        proc = MagicMock()
        proc.returncode = 1
        proc.stderr = "some error"

        with (
            patch("idfkit.simulation.expand.subprocess.run", return_value=proc),
            pytest.raises(ExpandObjectsError, match=r"did not produce expanded\.idf") as exc_info,
        ):
            expand_objects(model_with_hvac_template, energyplus=mock_config)
        assert exc_info.value.preprocessor == "ExpandObjects"
        assert exc_info.value.exit_code == 1
        assert exc_info.value.stderr == "some error"

    def test_timeout_raises(self, model_with_hvac_template: IDFDocument, mock_config: EnergyPlusConfig) -> None:
        import subprocess

        with (
            patch(
                "idfkit.simulation.expand.subprocess.run",
                side_effect=subprocess.TimeoutExpired(cmd="ExpandObjects", timeout=1),
            ),
            pytest.raises(ExpandObjectsError, match="timed out") as exc_info,
        ):
            expand_objects(model_with_hvac_template, energyplus=mock_config, timeout=1.0)
        assert exc_info.value.preprocessor == "ExpandObjects"
        assert exc_info.value.exit_code is None

    def test_os_error_raises(self, model_with_hvac_template: IDFDocument, mock_config: EnergyPlusConfig) -> None:
        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=OSError("Permission denied")),
            pytest.raises(ExpandObjectsError, match="Failed to start ExpandObjects") as exc_info,
        ):
            expand_objects(model_with_hvac_template, energyplus=mock_config)
        assert exc_info.value.preprocessor == "ExpandObjects"
        assert exc_info.value.exit_code is None
        assert exc_info.value.stderr is None

    def test_auto_discovers_energyplus(
        self, model_with_hvac_template: IDFDocument, mock_config: EnergyPlusConfig
    ) -> None:
        with (
            patch("idfkit.simulation.expand.find_energyplus", return_value=mock_config) as mock_find,
            patch("idfkit.simulation.expand.subprocess.run", side_effect=_fake_expand_run()),
        ):
            expand_objects(model_with_hvac_template)
        mock_find.assert_called_once()

    def test_document_expand_method(self, model_with_hvac_template: IDFDocument, mock_config: EnergyPlusConfig) -> None:
        with patch("idfkit.simulation.expand.subprocess.run", side_effect=_fake_expand_run()):
            expanded = model_with_hvac_template.expand(energyplus=mock_config)
        assert isinstance(expanded, IDFDocument)
        assert "Zone" in expanded

    def test_does_not_mutate_original(
        self, model_with_hvac_template: IDFDocument, mock_config: EnergyPlusConfig
    ) -> None:
        original_types = set(model_with_hvac_template.keys())

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=_fake_expand_run()):
            expand_objects(model_with_hvac_template, energyplus=mock_config)

        assert set(model_with_hvac_template.keys()) == original_types
        assert "HVACTemplate:Zone:IdealLoadsAirSystem" in model_with_hvac_template

    def test_copies_idd_to_run_dir(self, model_with_hvac_template: IDFDocument, mock_config: EnergyPlusConfig) -> None:
        idd_found: list[bool] = []

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            # Assert during execution, before the run_dir is cleaned up
            idd_found.append((cwd / "Energy+.idd").is_file())
            (cwd / "expanded.idf").write_text("Version, 24.1;\n\nZone,\n  Office;\n")
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run):
            expand_objects(model_with_hvac_template, energyplus=mock_config)

        assert len(idd_found) == 1
        assert idd_found[0] is True

    def test_skips_subprocess_when_nothing_to_expand(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("Zone", "Office", {"x_origin": 0.0})

        with patch("idfkit.simulation.expand.subprocess.run") as mock_run:
            result = expand_objects(doc)

        mock_run.assert_not_called()
        assert isinstance(result, IDFDocument)
        assert result is not doc


# ---------------------------------------------------------------------------
# Detection helper tests
# ---------------------------------------------------------------------------


class TestNeedsExpansion:
    def test_uses_schema_group(self, model_with_hvac_template: IDFDocument) -> None:
        assert model_with_hvac_template.schema is not None
        assert _needs_expansion(model_with_hvac_template) is True

    def test_ignores_ground_heat_transfer(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Control", "", {"name": "test", "run_basement_preprocessor": "No"}, validate=False)
        assert _needs_expansion(doc) is False

    def test_false_for_plain_model(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("Zone", "Office", {"x_origin": 0.0})
        assert _needs_expansion(doc) is False

    def test_falls_back_to_prefix_without_schema(self) -> None:
        doc = IDFDocument(version=(24, 1, 0), schema=None)
        doc.addidfobject(IDFObject(obj_type="HVACTemplate:Zone:IdealLoadsAirSystem", name="Test", data={}))
        assert _needs_expansion(doc) is True

    def test_prefix_fallback_negative(self) -> None:
        doc = IDFDocument(version=(24, 1, 0), schema=None)
        doc.addidfobject(IDFObject(obj_type="Zone", name="Office", data={}))
        assert _needs_expansion(doc) is False


class TestHasSlabObjects:
    def test_detects_slab_objects(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        assert _has_slab_objects(doc) is True

    def test_false_for_basement_only(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        assert _has_slab_objects(doc) is False

    def test_false_for_plain_model(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("Zone", "Office", {"x_origin": 0.0})
        assert _has_slab_objects(doc) is False

    def test_prefix_fallback(self) -> None:
        doc = IDFDocument(version=(24, 1, 0), schema=None)
        doc.addidfobject(IDFObject(obj_type="GroundHeatTransfer:Slab:Materials", name="", data={}))
        assert _has_slab_objects(doc) is True


class TestHasBasementObjects:
    def test_detects_basement_objects(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        assert _has_basement_objects(doc) is True

    def test_false_for_slab_only(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        assert _has_basement_objects(doc) is False

    def test_false_for_plain_model(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("Zone", "Office", {"x_origin": 0.0})
        assert _has_basement_objects(doc) is False


# ---------------------------------------------------------------------------
# needs_ground_heat_preprocessing tests
# ---------------------------------------------------------------------------


class TestNeedsGroundHeatPreprocessing:
    def test_true_for_slab_objects(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        assert needs_ground_heat_preprocessing(doc) is True

    def test_true_for_basement_objects(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        assert needs_ground_heat_preprocessing(doc) is True

    def test_true_for_both_slab_and_basement(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        assert needs_ground_heat_preprocessing(doc) is True

    def test_false_for_plain_model(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("Zone", "Office", {"x_origin": 0.0})
        assert needs_ground_heat_preprocessing(doc) is False

    def test_false_for_hvac_template_only(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("HVACTemplate:Zone:IdealLoadsAirSystem", "Test", {"zone_name": "Office"}, validate=False)
        assert needs_ground_heat_preprocessing(doc) is False


# ---------------------------------------------------------------------------
# Fatal preprocessor message detection tests
# ---------------------------------------------------------------------------


class TestCheckForFatalPreprocessorMessage:
    def _make_proc(self, returncode: int = 0) -> MagicMock:
        proc = MagicMock()
        proc.returncode = returncode
        proc.stderr = "STOP GroundTempCalc Terminated--Error(s) Detected.\n"
        return proc

    def test_raises_on_fatal_single_line(self, tmp_path: Path) -> None:
        output = tmp_path / "SLABSurfaceTemps.TXT"
        output.write_text(" Output:PreprocessorMessage,GroundTempCalc - Slab,Fatal,\nNo in.epw file found;\n")
        with pytest.raises(ExpandObjectsError, match="fatal error") as exc_info:
            _check_for_fatal_preprocessor_message(output, label="Slab", proc=self._make_proc())
        assert exc_info.value.preprocessor == "Slab"
        assert exc_info.value.exit_code == 0
        assert "No in.epw file found" in str(exc_info.value.stderr)

    def test_raises_on_fatal_inline(self, tmp_path: Path) -> None:
        output = tmp_path / "EPObjects.TXT"
        output.write_text("Output:PreprocessorMessage,GroundTempCalc - Basement,Fatal,No in.epw file found;\n")
        with pytest.raises(ExpandObjectsError, match="fatal error") as exc_info:
            _check_for_fatal_preprocessor_message(output, label="Basement", proc=self._make_proc())
        assert exc_info.value.preprocessor == "Basement"

    def test_no_raise_on_severe_warning(self, tmp_path: Path) -> None:
        output = tmp_path / "SLABSurfaceTemps.TXT"
        output.write_text(
            " Output:PreprocessorMessage,GroundTempCalc - Slab,Severe,\n"
            "The calculated ground temperature exceeds -100 C;\n\n"
            "SurfaceProperty:OtherSideCoefficients,\n"
            "  surfPropOthSdCoefSlabAverage;\n"
        )
        # Should NOT raise — Severe is a warning, not a fatal error
        _check_for_fatal_preprocessor_message(output, label="Slab", proc=self._make_proc())

    def test_no_raise_on_clean_output(self, tmp_path: Path) -> None:
        output = tmp_path / "SLABSurfaceTemps.TXT"
        output.write_text(
            "! This is clean output\nSurfaceProperty:OtherSideCoefficients,\n  surfPropOthSdCoefSlabAverage;\n"
        )
        _check_for_fatal_preprocessor_message(output, label="Slab", proc=self._make_proc())

    def test_preserves_exit_code(self, tmp_path: Path) -> None:
        output = tmp_path / "EPObjects.TXT"
        output.write_text("Output:PreprocessorMessage,GroundTempCalc - Basement,Fatal,Bad input;\n")
        with pytest.raises(ExpandObjectsError) as exc_info:
            _check_for_fatal_preprocessor_message(output, label="Basement", proc=self._make_proc(returncode=1))
        assert exc_info.value.exit_code == 1


class TestCheckProcessExitCode:
    def test_no_raise_on_zero(self) -> None:
        proc = MagicMock()
        proc.returncode = 0
        proc.stderr = ""
        _check_process_exit_code(proc, label="Slab")

    def test_raises_on_sigsegv(self) -> None:
        proc = MagicMock()
        proc.returncode = 139
        proc.stderr = "Segmentation fault\n"
        with pytest.raises(ExpandObjectsError, match="crashed with exit code 139") as exc_info:
            _check_process_exit_code(proc, label="Slab")
        assert exc_info.value.preprocessor == "Slab"
        assert exc_info.value.exit_code == 139
        assert "Segmentation fault" in str(exc_info.value.stderr)

    def test_raises_on_nonzero(self) -> None:
        proc = MagicMock()
        proc.returncode = 1
        proc.stderr = ""
        with pytest.raises(ExpandObjectsError, match="crashed with exit code 1") as exc_info:
            _check_process_exit_code(proc, label="Basement")
        assert exc_info.value.preprocessor == "Basement"
        assert exc_info.value.exit_code == 1


class TestCheckOutputNotEmpty:
    def test_no_raise_on_nonempty(self, tmp_path: Path) -> None:
        output = tmp_path / "SLABSurfaceTemps.TXT"
        output.write_text("SurfaceProperty:OtherSideCoefficients,\n  surfPropOthSdCoefSlabAverage;\n")
        proc = MagicMock()
        proc.returncode = 0
        proc.stderr = ""
        _check_output_not_empty(output, label="Slab", proc=proc)

    def test_raises_on_empty_file(self, tmp_path: Path) -> None:
        output = tmp_path / "EPObjects.TXT"
        output.write_text("")
        proc = MagicMock()
        proc.returncode = 0
        proc.stderr = "1-D solution has not converged\n"
        with pytest.raises(ExpandObjectsError, match="empty output file") as exc_info:
            _check_output_not_empty(output, label="Basement", proc=proc)
        assert exc_info.value.preprocessor == "Basement"
        assert exc_info.value.exit_code == 0
        assert "1-D solution has not converged" in str(exc_info.value.stderr)

    def test_preserves_exit_code(self, tmp_path: Path) -> None:
        output = tmp_path / "SLABSurfaceTemps.TXT"
        output.write_text("")
        proc = MagicMock()
        proc.returncode = 139
        proc.stderr = ""
        with pytest.raises(ExpandObjectsError) as exc_info:
            _check_output_not_empty(output, label="Slab", proc=proc)
        assert exc_info.value.exit_code == 139


class TestSlabFatalMessage:
    """Verify run_slab_preprocessor detects fatal preprocessor messages."""

    def test_raises_on_slab_fatal(self, mock_config: EnergyPlusConfig) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_slab_preprocessor": "Yes"}, validate=False)

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n")
                (cwd / "GHTIn.idf").write_text("! input\n")
            elif exe_name == "Slab":
                # Slab exits 0 but writes a Fatal message
                (cwd / "SLABSurfaceTemps.TXT").write_text(
                    " Output:PreprocessorMessage,GroundTempCalc - Slab,Fatal,\nNo in.epw file found;\n"
                )
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = "STOP GroundTempCalc Terminated--Error(s) Detected.\n"
            return result

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run),
            pytest.raises(ExpandObjectsError, match="fatal error") as exc_info,
        ):
            run_slab_preprocessor(doc, energyplus=mock_config)
        assert exc_info.value.preprocessor == "Slab"
        assert exc_info.value.exit_code == 0


class TestBasementFatalMessage:
    """Verify run_basement_preprocessor detects fatal preprocessor messages."""

    def test_raises_on_basement_fatal(self, mock_config: EnergyPlusConfig) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_basement_preprocessor": "Yes"}, validate=False)

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n")
                (cwd / "BasementGHTIn.idf").write_text("! input\n")
            elif exe_name == "Basement":
                (cwd / "EPObjects.TXT").write_text(
                    "Output:PreprocessorMessage,GroundTempCalc - Basement,Fatal,No in.epw file found;\n"
                )
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = "STOP GroundTempCalc Terminated--Error(s) Detected.\n"
            return result

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run),
            pytest.raises(ExpandObjectsError, match="fatal error") as exc_info,
        ):
            run_basement_preprocessor(doc, energyplus=mock_config)
        assert exc_info.value.preprocessor == "Basement"
        assert exc_info.value.exit_code == 0


# ---------------------------------------------------------------------------
# run_slab_preprocessor tests
# ---------------------------------------------------------------------------


class TestRunSlabPreprocessor:
    def test_skips_when_no_slab_objects(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("Zone", "Office", {"x_origin": 0.0})

        with patch("idfkit.simulation.expand.subprocess.run") as mock_run:
            result = run_slab_preprocessor(doc)

        mock_run.assert_not_called()
        assert result is not doc
        assert "Zone" in result

    def test_runs_expand_then_slab(self, mock_config: EnergyPlusConfig) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Slab:BldgProps", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_slab_preprocessor": "Yes"}, validate=False)
        doc.add("Zone", "Office", {"x_origin": 0.0})

        calls: list[str] = []

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            calls.append(exe_name)

            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n\nZone,\n  Office;\n")
                (cwd / "GHTIn.idf").write_text("! slab input extracted\n")
            elif exe_name == "Slab":
                (cwd / "SLABSurfaceTemps.TXT").write_text(
                    "Site:GroundTemperature:BuildingSurface,\n  18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18;\n"
                )

            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run):
            expanded = run_slab_preprocessor(doc, energyplus=mock_config)

        assert calls == ["ExpandObjects", "Slab"]
        assert "Site:GroundTemperature:BuildingSurface" in expanded

    def test_raises_when_ght_input_missing(self, mock_config: EnergyPlusConfig) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=_fake_expand_run()),
            pytest.raises(ExpandObjectsError, match=r"did not produce GHTIn\.idf") as exc_info,
        ):
            run_slab_preprocessor(doc, energyplus=mock_config)
        assert exc_info.value.preprocessor == "ExpandObjects"

    def test_raises_when_slab_exe_missing(self, tmp_path: Path) -> None:
        exe = tmp_path / "energyplus"
        exe.touch()
        idd = tmp_path / "Energy+.idd"
        idd.write_text("!IDD_Version 24.1.0\n")
        expand_exe = tmp_path / "ExpandObjects"
        expand_exe.touch()
        # No PreProcess/GrndTempCalc/Slab
        config = EnergyPlusConfig(executable=exe, version=(24, 1, 0), install_dir=tmp_path, idd_path=idd)

        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            (cwd / "expanded.idf").write_text("Version, 24.1;\n")
            (cwd / "GHTIn.idf").write_text("! input\n")
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run),
            pytest.raises(ExpandObjectsError, match="Slab preprocessor not found") as exc_info,
        ):
            run_slab_preprocessor(doc, energyplus=config)
        assert exc_info.value.preprocessor == "Slab"

    def test_copies_idds_to_run_dir(self, mock_config: EnergyPlusConfig) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)

        idd_copied: list[str] = []

        original_copy2 = shutil.copy2

        def track_copy(src: object, dst: object) -> object:
            idd_copied.append(str(dst))
            return original_copy2(src, dst)  # type: ignore[arg-type]

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n")
                (cwd / "GHTIn.idf").write_text("! input\n")
            elif exe_name == "Slab":
                (cwd / "SLABSurfaceTemps.TXT").write_text("! temps\n")
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run),
            patch("idfkit.simulation.expand.shutil.copy2", side_effect=track_copy),
        ):
            run_slab_preprocessor(doc, energyplus=mock_config)

        # Energy+.idd for ExpandObjects + SlabGHT.idd for Slab
        # (idd_copied is populated during execution, before cleanup)
        assert len(idd_copied) == 2
        idd_names = [Path(p).name for p in idd_copied]
        assert "Energy+.idd" in idd_names
        assert "SlabGHT.idd" in idd_names

    def test_weather_file_copied(self, mock_config: EnergyPlusConfig, tmp_path: Path) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)

        epw = tmp_path / "test.epw"
        epw.write_text("LOCATION,Chicago\n")

        epw_checks: list[tuple[bool, str]] = []

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            if exe_name == "ExpandObjects":
                # Check weather file during execution, before cleanup
                epw_path = cwd / "in.epw"
                epw_checks.append((epw_path.is_file(), epw_path.read_text() if epw_path.is_file() else ""))
                (cwd / "expanded.idf").write_text("Version, 24.1;\n")
                (cwd / "GHTIn.idf").write_text("! input\n")
            elif exe_name == "Slab":
                (cwd / "SLABSurfaceTemps.TXT").write_text("! temps\n")
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run):
            run_slab_preprocessor(doc, energyplus=mock_config, weather=epw)

        assert len(epw_checks) == 1
        assert epw_checks[0][0] is True
        assert epw_checks[0][1] == "LOCATION,Chicago\n"

    def test_raises_on_slab_crash(self, mock_config: EnergyPlusConfig) -> None:
        """Slab SIGSEGV (exit code 139) with empty output files is caught."""
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_slab_preprocessor": "Yes"}, validate=False)

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n")
                (cwd / "GHTIn.idf").write_text("! input\n")
            elif exe_name == "Slab":
                # Simulate SIGSEGV: empty output file, non-zero exit code
                (cwd / "SLABSurfaceTemps.TXT").write_text("")
            result = MagicMock()
            result.returncode = 0 if exe_name == "ExpandObjects" else 139
            result.stdout = ""
            result.stderr = "Segmentation fault\n" if exe_name == "Slab" else ""
            return result

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run),
            pytest.raises(ExpandObjectsError, match="crashed with exit code 139") as exc_info,
        ):
            run_slab_preprocessor(doc, energyplus=mock_config)
        assert exc_info.value.preprocessor == "Slab"
        assert exc_info.value.exit_code == 139

    def test_raises_on_empty_slab_output(self, mock_config: EnergyPlusConfig) -> None:
        """Slab solver failure producing empty output file is caught."""
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_slab_preprocessor": "Yes"}, validate=False)

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n")
                (cwd / "GHTIn.idf").write_text("! input\n")
            elif exe_name == "Slab":
                # Solver failure: exit code 0, but empty output
                (cwd / "SLABSurfaceTemps.TXT").write_text("")
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run),
            pytest.raises(ExpandObjectsError, match="empty output file") as exc_info,
        ):
            run_slab_preprocessor(doc, energyplus=mock_config)
        assert exc_info.value.preprocessor == "Slab"


# ---------------------------------------------------------------------------
# run_basement_preprocessor tests
# ---------------------------------------------------------------------------


class TestRunBasementPreprocessor:
    def test_skips_when_no_basement_objects(self) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("Zone", "Office", {"x_origin": 0.0})

        with patch("idfkit.simulation.expand.subprocess.run") as mock_run:
            result = run_basement_preprocessor(doc)

        mock_run.assert_not_called()
        assert result is not doc

    def test_runs_expand_then_basement(self, mock_config: EnergyPlusConfig) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_basement_preprocessor": "Yes"}, validate=False)
        doc.add("Zone", "Office", {"x_origin": 0.0})

        calls: list[str] = []

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            calls.append(exe_name)

            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n\nZone,\n  Office;\n")
                (cwd / "BasementGHTIn.idf").write_text("! basement input extracted\n")
            elif exe_name == "Basement":
                (cwd / "EPObjects.TXT").write_text(
                    "Site:GroundTemperature:BuildingSurface,\n  16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16;\n"
                )

            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run):
            expanded = run_basement_preprocessor(doc, energyplus=mock_config)

        assert calls == ["ExpandObjects", "Basement"]
        assert "Site:GroundTemperature:BuildingSurface" in expanded

    def test_raises_when_basement_input_missing(self, mock_config: EnergyPlusConfig) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=_fake_expand_run()),
            pytest.raises(ExpandObjectsError, match=r"did not produce BasementGHTIn\.idf") as exc_info,
        ):
            run_basement_preprocessor(doc, energyplus=mock_config)
        assert exc_info.value.preprocessor == "ExpandObjects"

    def test_weather_file_copied(self, mock_config: EnergyPlusConfig, tmp_path: Path) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_basement_preprocessor": "Yes"}, validate=False)
        doc.add("Zone", "Office", {"x_origin": 0.0})

        epw = tmp_path / "test.epw"
        epw.write_text("LOCATION,Chicago\n")

        epw_checks: list[bool] = []

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            if exe_name == "ExpandObjects":
                # Check weather file during execution, before cleanup
                epw_checks.append((cwd / "in.epw").is_file())
                (cwd / "expanded.idf").write_text("Version, 24.1;\n\nZone,\n  Office;\n")
                (cwd / "BasementGHTIn.idf").write_text("! basement input extracted\n")
            elif exe_name == "Basement":
                (cwd / "EPObjects.TXT").write_text(
                    "Site:GroundTemperature:BuildingSurface,\n  16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16;\n"
                )
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run):
            expanded = run_basement_preprocessor(doc, energyplus=mock_config, weather=epw)

        assert len(epw_checks) == 1
        assert epw_checks[0] is True
        assert "Site:GroundTemperature:BuildingSurface" in expanded

    def test_raises_when_weather_file_missing(self, mock_config: EnergyPlusConfig, tmp_path: Path) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)

        with pytest.raises(ExpandObjectsError, match="Weather file not found"):
            run_basement_preprocessor(doc, energyplus=mock_config, weather=tmp_path / "nonexistent.epw")

    def test_raises_on_basement_crash(self, mock_config: EnergyPlusConfig) -> None:
        """Basement crash (non-zero exit code) is caught."""
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_basement_preprocessor": "Yes"}, validate=False)

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n")
                (cwd / "BasementGHTIn.idf").write_text("! input\n")
            elif exe_name == "Basement":
                (cwd / "EPObjects.TXT").write_text("")
            result = MagicMock()
            result.returncode = 0 if exe_name == "ExpandObjects" else 134
            result.stdout = ""
            result.stderr = "Aborted\n" if exe_name == "Basement" else ""
            return result

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run),
            pytest.raises(ExpandObjectsError, match="crashed with exit code 134") as exc_info,
        ):
            run_basement_preprocessor(doc, energyplus=mock_config)
        assert exc_info.value.preprocessor == "Basement"
        assert exc_info.value.exit_code == 134

    def test_raises_on_empty_basement_output(self, mock_config: EnergyPlusConfig) -> None:
        """Basement solver non-convergence producing empty EPObjects.TXT is caught."""
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_basement_preprocessor": "Yes"}, validate=False)

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n")
                (cwd / "BasementGHTIn.idf").write_text("! input\n")
            elif exe_name == "Basement":
                # Solver non-convergence: exit 0, empty output
                (cwd / "EPObjects.TXT").write_text("")
            result = MagicMock()
            result.returncode = 0
            result.stdout = "1-D solution has not converged\nIYR=MAXYR-1, PROGRAM TERMINATED\n"
            result.stderr = ""
            return result

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run),
            pytest.raises(ExpandObjectsError, match="empty output file") as exc_info,
        ):
            run_basement_preprocessor(doc, energyplus=mock_config)
        assert exc_info.value.preprocessor == "Basement"
        assert exc_info.value.exit_code == 0

    def test_raises_when_basement_exe_missing(self, tmp_path: Path) -> None:
        exe = tmp_path / "energyplus"
        exe.touch()
        idd = tmp_path / "Energy+.idd"
        idd.write_text("!IDD_Version 24.1.0\n")
        expand_exe = tmp_path / "ExpandObjects"
        expand_exe.touch()
        config = EnergyPlusConfig(executable=exe, version=(24, 1, 0), install_dir=tmp_path, idd_path=idd)

        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            (cwd / "expanded.idf").write_text("Version, 24.1;\n")
            (cwd / "BasementGHTIn.idf").write_text("! input\n")
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with (
            patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run),
            pytest.raises(ExpandObjectsError, match="Basement preprocessor not found") as exc_info,
        ):
            run_basement_preprocessor(doc, energyplus=config)
        assert exc_info.value.preprocessor == "Basement"


# ---------------------------------------------------------------------------
# run_preprocessing tests
# ---------------------------------------------------------------------------


class TestRunPreprocessing:
    """Tests for run_preprocessing used by simulate()."""

    def test_runs_slab_when_ght_input_produced(self, mock_config: EnergyPlusConfig, tmp_path: Path) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_slab_preprocessor": "Yes"}, validate=False)
        doc.add("Zone", "Office", {"x_origin": 0.0})

        epw = tmp_path / "weather.epw"
        epw.write_text("LOCATION,Chicago\n")

        calls: list[str] = []

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            calls.append(exe_name)
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n\nZone,\n  Office;\n")
                (cwd / "GHTIn.idf").write_text("! slab input\n")
            elif exe_name == "Slab":
                (cwd / "SLABSurfaceTemps.TXT").write_text(
                    "Site:GroundTemperature:BuildingSurface,\n  18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18;\n"
                )
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run):
            expanded = run_preprocessing(doc, energyplus=mock_config, weather=epw, timeout=120.0)

        assert calls == ["ExpandObjects", "Slab"]
        assert "Site:GroundTemperature:BuildingSurface" in expanded

    def test_runs_basement_when_ght_input_produced(self, mock_config: EnergyPlusConfig, tmp_path: Path) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_basement_preprocessor": "Yes"}, validate=False)
        doc.add("Zone", "Office", {"x_origin": 0.0})

        epw = tmp_path / "weather.epw"
        epw.write_text("LOCATION,Chicago\n")

        calls: list[str] = []

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            calls.append(exe_name)
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n\nZone,\n  Office;\n")
                (cwd / "BasementGHTIn.idf").write_text("! basement input\n")
            elif exe_name == "Basement":
                (cwd / "EPObjects.TXT").write_text(
                    "Site:GroundTemperature:BuildingSurface,\n  16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16;\n"
                )
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run):
            expanded = run_preprocessing(doc, energyplus=mock_config, weather=epw, timeout=120.0)

        assert calls == ["ExpandObjects", "Basement"]
        assert "Site:GroundTemperature:BuildingSurface" in expanded

    def test_runs_both_slab_and_basement(self, mock_config: EnergyPlusConfig, tmp_path: Path) -> None:
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Basement:SimParameters", "", {}, validate=False)
        doc.add(
            "GroundHeatTransfer:Control",
            "",
            {"run_slab_preprocessor": "Yes", "run_basement_preprocessor": "Yes"},
            validate=False,
        )
        doc.add("Zone", "Office", {"x_origin": 0.0})

        epw = tmp_path / "weather.epw"
        epw.write_text("LOCATION,Chicago\n")

        calls: list[str] = []

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            calls.append(exe_name)
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n\nZone,\n  Office;\n")
                (cwd / "GHTIn.idf").write_text("! slab input\n")
                (cwd / "BasementGHTIn.idf").write_text("! basement input\n")
            elif exe_name == "Slab":
                (cwd / "SLABSurfaceTemps.TXT").write_text("! slab temps\n")
            elif exe_name == "Basement":
                (cwd / "EPObjects.TXT").write_text("! basement temps\n")
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run):
            run_preprocessing(doc, energyplus=mock_config, weather=epw, timeout=120.0)

        assert calls == ["ExpandObjects", "Slab", "Basement"]

    def test_skips_slab_when_ght_input_not_produced(self, mock_config: EnergyPlusConfig, tmp_path: Path) -> None:
        """When ExpandObjects doesn't produce GHTIn.idf (control=No), Slab is skipped."""
        doc = new_document(version=(24, 1, 0))
        doc.add("GroundHeatTransfer:Slab:Materials", "", {}, validate=False)
        doc.add("GroundHeatTransfer:Control", "", {"run_slab_preprocessor": "No"}, validate=False)
        doc.add("Zone", "Office", {"x_origin": 0.0})

        epw = tmp_path / "weather.epw"
        epw.write_text("LOCATION,Chicago\n")

        calls: list[str] = []

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            calls.append(exe_name)
            if exe_name == "ExpandObjects":
                # No GHTIn.idf produced (run_slab_preprocessor=No)
                (cwd / "expanded.idf").write_text("Version, 24.1;\n\nZone,\n  Office;\n")
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run):
            run_preprocessing(doc, energyplus=mock_config, weather=epw, timeout=120.0)

        assert calls == ["ExpandObjects"]

    def test_expand_objects_only_when_no_ght_inputs(self, mock_config: EnergyPlusConfig, tmp_path: Path) -> None:
        """When model has HVACTemplate but no GHT objects, only ExpandObjects runs."""
        doc = new_document(version=(24, 1, 0))
        doc.add("HVACTemplate:Zone:IdealLoadsAirSystem", "Test", {"zone_name": "Office"}, validate=False)
        doc.add("Zone", "Office", {"x_origin": 0.0})

        epw = tmp_path / "weather.epw"
        epw.write_text("LOCATION,Chicago\n")

        calls: list[str] = []

        def fake_run(cmd: list[str], **kwargs: object) -> MagicMock:
            cwd = Path(str(kwargs.get("cwd", "")))
            exe_name = Path(cmd[0]).name
            calls.append(exe_name)
            if exe_name == "ExpandObjects":
                (cwd / "expanded.idf").write_text("Version, 24.1;\n\nZone,\n  Office;\n")
            result = MagicMock()
            result.returncode = 0
            result.stdout = ""
            result.stderr = ""
            return result

        with patch("idfkit.simulation.expand.subprocess.run", side_effect=fake_run):
            run_preprocessing(doc, energyplus=mock_config, weather=epw, timeout=120.0)

        assert calls == ["ExpandObjects"]


# ---------------------------------------------------------------------------
# EnergyPlusConfig property tests
# ---------------------------------------------------------------------------


class TestConfigProperties:
    def test_slab_exe(self, mock_config: EnergyPlusConfig) -> None:
        assert mock_config.slab_exe is not None
        assert mock_config.slab_exe.name == "Slab"

    def test_slab_idd(self, mock_config: EnergyPlusConfig) -> None:
        assert mock_config.slab_idd is not None
        assert mock_config.slab_idd.name == "SlabGHT.idd"

    def test_basement_exe(self, mock_config: EnergyPlusConfig) -> None:
        assert mock_config.basement_exe is not None
        assert mock_config.basement_exe.name == "Basement"

    def test_basement_idd(self, mock_config: EnergyPlusConfig) -> None:
        assert mock_config.basement_idd is not None
        assert mock_config.basement_idd.name == "BasementGHT.idd"

    def test_missing_preprocessors(self, tmp_path: Path) -> None:
        exe = tmp_path / "energyplus"
        exe.touch()
        idd = tmp_path / "Energy+.idd"
        idd.write_text("!IDD_Version 24.1.0\n")
        config = EnergyPlusConfig(executable=exe, version=(24, 1, 0), install_dir=tmp_path, idd_path=idd)

        assert config.slab_exe is None
        assert config.slab_idd is None
        assert config.basement_exe is None
        assert config.basement_idd is None



"""Tests for the async EnergyPlus simulation runner and batch modules."""

from __future__ import annotations

import asyncio
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from conftest import InMemoryAsyncFileSystem, InMemoryFileSystem

from idfkit import new_document
from idfkit.exceptions import SimulationError
from idfkit.simulation.async_batch import SimulationEvent, async_simulate_batch, async_simulate_batch_stream
from idfkit.simulation.async_runner import async_simulate
from idfkit.simulation.batch import SimulationJob
from idfkit.simulation.cache import SimulationCache
from idfkit.simulation.config import EnergyPlusConfig


@pytest.fixture
def mock_config(tmp_path: Path) -> EnergyPlusConfig:
    """Create a mock EnergyPlusConfig pointing to tmp_path."""
    exe = tmp_path / "energyplus"
    exe.touch()
    exe.chmod(0o755)
    idd = tmp_path / "Energy+.idd"
    idd.write_text("!IDD_Version 24.1.0\n")

    # ExpandObjects
    expand_exe = tmp_path / "ExpandObjects"
    expand_exe.touch()
    expand_exe.chmod(0o755)

    # Slab & Basement preprocessors
    preprocess = tmp_path / "PreProcess" / "GrndTempCalc"
    preprocess.mkdir(parents=True)
    slab_exe = preprocess / "Slab"
    slab_exe.touch()
    slab_exe.chmod(0o755)
    (preprocess / "SlabGHT.idd").write_text("! Slab IDD\n")
    basement_exe = preprocess / "Basement"
    basement_exe.touch()
    basement_exe.chmod(0o755)
    (preprocess / "BasementGHT.idd").write_text("! Basement IDD\n")

    return EnergyPlusConfig(
        executable=exe,
        version=(24, 1, 0),
        install_dir=tmp_path,
        idd_path=idd,
    )


@pytest.fixture
def weather_file(tmp_path: Path) -> Path:
    """Create a dummy weather file."""
    epw = tmp_path / "weather.epw"
    epw.write_text("LOCATION,Chicago\n")
    return epw


def _make_mock_process(returncode: int = 0, stdout: bytes = b"ok", stderr: bytes = b"") -> MagicMock:
    """Create a mock asyncio.subprocess.Process."""
    proc = MagicMock()
    proc.communicate = AsyncMock(return_value=(stdout, stderr))
    proc.returncode = returncode
    proc.kill = MagicMock()
    proc.wait = AsyncMock()
    return proc


# ---------------------------------------------------------------------------
# async_simulate
# ---------------------------------------------------------------------------


class TestAsyncSimulate:
    """Tests for async_simulate()."""

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_success(self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path) -> None:
        mock_exec.return_value = _make_mock_process(returncode=0)
        model = new_document()
        result = await async_simulate(model, weather_file, energyplus=mock_config)
        assert result.success
        assert result.exit_code == 0
        mock_exec.assert_called_once()

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_failure(self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path) -> None:
        mock_exec.return_value = _make_mock_process(returncode=1, stderr=b"error occurred")
        model = new_document()
        result = await async_simulate(model, weather_file, energyplus=mock_config)
        assert not result.success
        assert result.exit_code == 1

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_timeout(self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path) -> None:
        proc = _make_mock_process()
        proc.communicate = AsyncMock(side_effect=asyncio.TimeoutError)
        mock_exec.return_value = proc
        model = new_document()

        with pytest.raises(SimulationError, match="timed out"):
            await async_simulate(model, weather_file, energyplus=mock_config, timeout=10)

        proc.kill.assert_called_once()

    @pytest.mark.asyncio
    async def test_missing_weather(self, mock_config: EnergyPlusConfig) -> None:
        model = new_document()
        with pytest.raises(SimulationError, match="Weather file not found"):
            await async_simulate(model, "/nonexistent/weather.epw", energyplus=mock_config)

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_model_not_mutated(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        mock_exec.return_value = _make_mock_process()
        model = new_document()
        assert "Output:SQLite" not in model
        await async_simulate(model, weather_file, energyplus=mock_config)
        assert "Output:SQLite" not in model

    @pytest.mark.asyncio
    async def test_fs_requires_output_dir(self, mock_config: EnergyPlusConfig, weather_file: Path) -> None:
        fs = InMemoryFileSystem()
        model = new_document()
        with pytest.raises(ValueError, match="output_dir is required"):
            await async_simulate(model, weather_file, energyplus=mock_config, fs=fs)

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_simulate_with_fs_uploads_results(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        mock_exec.return_value = _make_mock_process()
        fs = InMemoryFileSystem()
        model = new_document()
        result = await async_simulate(
            model,
            weather_file,
            energyplus=mock_config,
            output_dir="remote/output",
            fs=fs,
        )
        assert result.success
        assert result.fs is fs
        assert result.run_dir == Path("remote/output")
        uploaded = [k for k in fs._files if k.startswith("remote/output/")]
        assert len(uploaded) > 0

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_stdout_stderr_captured(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        mock_exec.return_value = _make_mock_process(returncode=0, stdout=b"simulation output", stderr=b"some warnings")
        model = new_document()
        result = await async_simulate(model, weather_file, energyplus=mock_config)
        assert result.stdout == "simulation output"
        assert result.stderr == "some warnings"

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_runtime_tracked(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        mock_exec.return_value = _make_mock_process()
        model = new_document()
        result = await async_simulate(model, weather_file, energyplus=mock_config)
        assert result.runtime_seconds >= 0


# ---------------------------------------------------------------------------
# async_simulate_batch
# ---------------------------------------------------------------------------


class TestAsyncSimulateBatch:
    """Tests for async_simulate_batch()."""

    @pytest.mark.asyncio
    async def test_empty_jobs_raises(self) -> None:
        with pytest.raises(ValueError, match="jobs must not be empty"):
            await async_simulate_batch([])

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_single_job(self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path) -> None:
        mock_exec.return_value = _make_mock_process()
        model = new_document()
        job = SimulationJob(model=model, weather=weather_file, label="single")
        result = await async_simulate_batch([job], energyplus=mock_config)
        assert len(result) == 1
        assert result[0].success
        assert result.all_succeeded

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_multiple_jobs(self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path) -> None:
        mock_exec.return_value = _make_mock_process()
        jobs = [SimulationJob(model=new_document(), weather=weather_file, label=f"job-{i}") for i in range(3)]
        result = await async_simulate_batch(jobs, energyplus=mock_config)
        assert len(result) == 3
        assert result.all_succeeded

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_preserves_order(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        """Results should be in the same order as input jobs."""
        call_count = 0

        async def make_proc(*_args: object, **_kwargs: object) -> MagicMock:
            nonlocal call_count
            call_count += 1
            return _make_mock_process(stdout=f"run-{call_count}".encode())

        mock_exec.side_effect = make_proc
        jobs = [SimulationJob(model=new_document(), weather=weather_file, label=f"job-{i}") for i in range(3)]
        result = await async_simulate_batch(jobs, energyplus=mock_config, max_concurrent=1)
        assert result[0].stdout == "run-1"
        assert result[1].stdout == "run-2"
        assert result[2].stdout == "run-3"

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_failed_job_does_not_cancel_others(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        call_count = 0

        async def make_proc(*_args: object, **_kwargs: object) -> MagicMock:
            nonlocal call_count
            call_count += 1
            if call_count == 2:
                return _make_mock_process(returncode=1, stderr=b"failed")
            return _make_mock_process()

        mock_exec.side_effect = make_proc
        jobs = [SimulationJob(model=new_document(), weather=weather_file, label=f"job-{i}") for i in range(3)]
        result = await async_simulate_batch(jobs, energyplus=mock_config, max_concurrent=1)
        assert len(result) == 3
        assert not result.all_succeeded
        assert len(result.succeeded) == 2
        assert len(result.failed) == 1

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_total_runtime_populated(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        mock_exec.return_value = _make_mock_process()
        job = SimulationJob(model=new_document(), weather=weather_file)
        result = await async_simulate_batch([job], energyplus=mock_config)
        assert result.total_runtime_seconds >= 0

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_cache_integration(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path, tmp_path: Path
    ) -> None:
        mock_exec.return_value = _make_mock_process()
        cache = SimulationCache(cache_dir=tmp_path / "cache")
        model = new_document()
        job = SimulationJob(model=model, weather=weather_file, label="cached")

        # First run — miss, then store
        result1 = await async_simulate_batch([job], energyplus=mock_config, cache=cache)
        assert result1[0].success
        assert mock_exec.call_count == 1

        # Second run — cache hit, subprocess not called again
        result2 = await async_simulate_batch([job], energyplus=mock_config, cache=cache)
        assert result2[0].success
        assert mock_exec.call_count == 1  # no additional call


# ---------------------------------------------------------------------------
# async_simulate_batch_stream
# ---------------------------------------------------------------------------


class TestAsyncSimulateBatchStream:
    """Tests for async_simulate_batch_stream()."""

    @pytest.mark.asyncio
    async def test_empty_jobs_raises(self) -> None:
        with pytest.raises(ValueError, match="jobs must not be empty"):
            async for _ in async_simulate_batch_stream([]):
                pass  # pragma: no cover

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_yields_all_events(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        mock_exec.return_value = _make_mock_process()
        jobs = [SimulationJob(model=new_document(), weather=weather_file, label=f"job-{i}") for i in range(3)]
        events: list[SimulationEvent] = []
        async for event in async_simulate_batch_stream(jobs, energyplus=mock_config):
            events.append(event)

        assert len(events) == 3
        # Each event should have correct total
        assert all(e.total == 3 for e in events)
        # Completed counts should reach 3
        completed_values = sorted(e.completed for e in events)
        assert completed_values == [1, 2, 3]
        # All indices should be present
        assert sorted(e.index for e in events) == [0, 1, 2]

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_event_contains_result(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        mock_exec.return_value = _make_mock_process()
        job = SimulationJob(model=new_document(), weather=weather_file, label="test-stream")
        events: list[SimulationEvent] = []
        async for event in async_simulate_batch_stream([job], energyplus=mock_config):
            events.append(event)

        assert len(events) == 1
        assert events[0].label == "test-stream"
        assert events[0].result.success
        assert events[0].index == 0
        assert events[0].completed == 1
        assert events[0].total == 1

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_early_break_cancels_remaining(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        """Breaking out of the stream should cancel remaining tasks."""
        call_count = 0

        async def slow_proc(*_args: object, **_kwargs: object) -> MagicMock:
            nonlocal call_count
            call_count += 1
            proc = _make_mock_process()
            if call_count > 1:
                # Make subsequent calls slow so they're still pending when we break
                original_communicate = proc.communicate

                async def slow_communicate() -> tuple[bytes, bytes]:
                    await asyncio.sleep(10)
                    return await original_communicate()

                proc.communicate = slow_communicate
            return proc

        mock_exec.side_effect = slow_proc
        jobs = [SimulationJob(model=new_document(), weather=weather_file, label=f"job-{i}") for i in range(5)]

        events: list[SimulationEvent] = []
        async for event in async_simulate_batch_stream(jobs, energyplus=mock_config, max_concurrent=1):
            events.append(event)
            break  # Only take the first result

        # Should have gotten exactly one event
        assert len(events) == 1


# ---------------------------------------------------------------------------
# SimulationEvent
# ---------------------------------------------------------------------------


class TestSimulationEvent:
    """Tests for SimulationEvent dataclass."""

    def test_frozen(self, tmp_path: Path) -> None:
        result = MagicMock()
        event = SimulationEvent(index=0, label="test", result=result, completed=1, total=1)
        with pytest.raises(AttributeError):
            event.label = "changed"  # type: ignore[misc]

    def test_fields(self, tmp_path: Path) -> None:
        result = MagicMock()
        event = SimulationEvent(index=2, label="my-sim", result=result, completed=3, total=5)
        assert event.index == 2
        assert event.label == "my-sim"
        assert event.result is result
        assert event.completed == 3
        assert event.total == 5


# ---------------------------------------------------------------------------
# AsyncFileSystem integration
# ---------------------------------------------------------------------------


class TestAsyncFileSystem:
    """Tests for async file system support in async_simulate()."""

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_simulate_with_async_fs_uploads_results(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        """async_simulate with an AsyncFileSystem should upload without blocking."""
        mock_exec.return_value = _make_mock_process()
        fs = InMemoryAsyncFileSystem()
        model = new_document()
        result = await async_simulate(
            model,
            weather_file,
            energyplus=mock_config,
            output_dir="remote/output",
            fs=fs,
        )
        assert result.success
        assert result.async_fs is fs
        assert result.fs is None
        assert result.run_dir == Path("remote/output")
        uploaded = [k for k in fs._files if k.startswith("remote/output/")]
        assert len(uploaded) > 0

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_simulate_with_sync_fs_wraps_in_thread(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        """async_simulate with a sync FileSystem should wrap upload in to_thread."""
        mock_exec.return_value = _make_mock_process()
        fs = InMemoryFileSystem()
        model = new_document()
        result = await async_simulate(
            model,
            weather_file,
            energyplus=mock_config,
            output_dir="remote/output",
            fs=fs,
        )
        assert result.success
        assert result.fs is fs
        assert result.async_fs is None
        assert result.run_dir == Path("remote/output")
        uploaded = [k for k in fs._files if k.startswith("remote/output/")]
        assert len(uploaded) > 0

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_async_fs_result_has_async_accessors(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        """SimulationResult from async fs should support async_errors() etc."""
        mock_exec.return_value = _make_mock_process()
        fs = InMemoryAsyncFileSystem()
        model = new_document()
        result = await async_simulate(
            model,
            weather_file,
            energyplus=mock_config,
            output_dir="remote/output",
            fs=fs,
        )
        # async_errors should not raise (even if no .err file exists)
        errors = await result.async_errors()
        assert errors is not None

    @pytest.mark.asyncio
    @patch("idfkit.simulation.async_runner.asyncio.create_subprocess_exec")
    async def test_async_batch_with_async_fs(
        self, mock_exec: AsyncMock, mock_config: EnergyPlusConfig, weather_file: Path
    ) -> None:
        """async_simulate_batch should accept an AsyncFileSystem."""
        mock_exec.return_value = _make_mock_process()
        fs = InMemoryAsyncFileSystem()
        jobs = [
            SimulationJob(model=new_document(), weather=weather_file, label="job-0", output_dir="batch/run-0"),
            SimulationJob(model=new_document(), weather=weather_file, label="job-1", output_dir="batch/run-1"),
        ]
        result = await async_simulate_batch(jobs, energyplus=mock_config, fs=fs)
        assert len(result) == 2
        assert result.all_succeeded
        # Both should have used the async fs
        for r in result.results:
            assert r.async_fs is fs

    @pytest.mark.asyncio
    async def test_sync_property_raises_when_only_async_fs_set(self) -> None:
        """Accessing sync properties on an async_fs-only result should raise."""
        from idfkit.simulation.result import SimulationResult

        fs = InMemoryAsyncFileSystem()
        result = SimulationResult(
            run_dir=Path("remote/output"),
            success=True,
            exit_code=0,
            stdout="",
            stderr="",
            runtime_seconds=0.0,
            async_fs=fs,
        )
        with pytest.raises(RuntimeError, match="async accessors"):
            _ = result.errors
        with pytest.raises(RuntimeError, match="async accessors"):
            _ = result.sql

    @pytest.mark.asyncio
    async def test_async_errors_with_actual_data(self) -> None:
        """async_errors should parse .err content from an async fs."""
        from idfkit.simulation.result import SimulationResult

        fs = InMemoryAsyncFileSystem()
        await fs.write_text(
            "run/eplusout.err",
            "Program Version,EnergyPlus, 24.1\n   ************* EnergyPlus Completed Successfully.\n",
        )
        result = SimulationResult(
            run_dir=Path("run"),
            success=True,
            exit_code=0,
            stdout="",
            stderr="",
            runtime_seconds=0.0,
            async_fs=fs,
        )
        errors = await result.async_errors()
        assert "completed successfully" in errors.summary()

    @pytest.mark.asyncio
    async def test_async_csv_with_actual_data(self) -> None:
        """async_csv should parse .csv content from an async fs."""
        from idfkit.simulation.result import SimulationResult

        fs = InMemoryAsyncFileSystem()
        csv_content = "Date/Time,Zone Mean Air Temperature [C](TimeStep)\n 01/01  01:00:00,21.5\n"
        await fs.write_text("run/eplusout.csv", csv_content)
        result = SimulationResult(
            run_dir=Path("run"),
            success=True,
            exit_code=0,
            stdout="",
            stderr="",
            runtime_seconds=0.0,
            async_fs=fs,
        )
        csv_result = await result.async_csv()
        assert csv_result is not None
        assert len(csv_result.columns) > 0

    @pytest.mark.asyncio
    async def test_async_variables_with_actual_data(self) -> None:
        """async_variables should parse .rdd content from an async fs."""
        from idfkit.simulation.result import SimulationResult

        fs = InMemoryAsyncFileSystem()
        rdd_content = "! Program Version,EnergyPlus, 24.1\nOutput:Variable,*,Zone Mean Air Temperature,hourly; !- [C]\n"
        await fs.write_text("run/eplusout.rdd", rdd_content)
        result = SimulationResult(
            run_dir=Path("run"),
            success=True,
            exit_code=0,
            stdout="",
            stderr="",
            runtime_seconds=0.0,
            async_fs=fs,
        )
        variables = await result.async_variables()
        assert variables is not None
        assert len(variables.variables) == 1

    @pytest.mark.asyncio
    async def test_async_html_with_actual_data(self) -> None:
        """async_html should parse HTML content from an async fs."""
        from idfkit.simulation.result import SimulationResult

        fs = InMemoryAsyncFileSystem()
        html_content = (
            "<html><body>"
            "<b>Report:</b><b>Test Report</b>"
            "<table><tr><td>Zone</td><td>Value</td></tr>"
            "<tr><td>Zone1</td><td>21.5</td></tr></table>"
            "</body></html>"
        )
        await fs.write_text("run/eplusoutTable.html", html_content)
        result = SimulationResult(
            run_dir=Path("run"),
            success=True,
            exit_code=0,
            stdout="",
            stderr="",
            runtime_seconds=0.0,
            async_fs=fs,
        )
        html_result = await result.async_html()
        assert html_result is not None

    @pytest.mark.asyncio
    async def test_from_directory_with_async_fs(self) -> None:
        """from_directory should accept async_fs and populate it on the result."""
        from idfkit.simulation.result import SimulationResult

        fs = InMemoryAsyncFileSystem()
        await fs.write_text(
            "run/eplusout.err",
            "Program Version,EnergyPlus, 24.1\n   ************* EnergyPlus Completed Successfully.\n",
        )
        result = SimulationResult.from_directory("run", async_fs=fs)
        assert result.async_fs is fs
        assert result.fs is None
        errors = await result.async_errors()
        assert "completed successfully" in errors.summary()

    def test_both_fs_and_async_fs_raises(self) -> None:
        """Setting both fs and async_fs should raise ValueError."""
        from idfkit.simulation.result import SimulationResult

        with pytest.raises(ValueError, match="mutually exclusive"):
            SimulationResult(
                run_dir=Path("run"),
                success=True,
                exit_code=0,
                stdout="",
                stderr="",
                runtime_seconds=0.0,
                fs=InMemoryFileSystem(),
                async_fs=InMemoryAsyncFileSystem(),
            )



from __future__ import annotations

from idfkit import IDFObject
from idfkit.simulation import FileSystem
from pathlib import Path


# --8<-- [start:example]
class ScheduleFileCache:
    """Cache for Schedule:File CSV data."""

    _cache: dict[str, list[float]]  # file_path -> values

    def get_values(
        self,
        obj: IDFObject,
        fs: FileSystem,
        base_path: Path,
    ) -> list[float]:
        """Get cached values or read from file."""


# --8<-- [end:example]



"""
Thermal property calculations for EnergyPlus constructions.

Provides R-value, U-value, and SHGC calculations for opaque and glazing
constructions based on material layer properties.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from .gas import typical_gap_r_value

if TYPE_CHECKING:
    from ..objects import IDFObject

# Construction layer field names in order (outside to inside)
_LAYER_FIELDS = (
    "outside_layer",
    "layer_2",
    "layer_3",
    "layer_4",
    "layer_5",
    "layer_6",
    "layer_7",
    "layer_8",
    "layer_9",
    "layer_10",
)

# Material types to search when resolving layer references
_MATERIAL_TYPES = (
    "Material",
    "Material:NoMass",
    "Material:AirGap",
    "WindowMaterial:Glazing",
    "WindowMaterial:Gas",
    "WindowMaterial:SimpleGlazingSystem",
)

# Standard surface film resistances (m²·K/W)
# From ASHRAE Fundamentals, Chapter 26
FILM_RESISTANCE = {
    # Exterior surface (15 mph wind)
    "exterior": 0.030,
    # Interior surface (still air)
    "interior": 0.120,
    # Interior surface - horizontal, heat flow up (ceiling)
    "interior_ceiling": 0.107,
    # Interior surface - horizontal, heat flow down (floor)
    "interior_floor": 0.160,
}

# NFRC standard film coefficients for windows (W/m²·K)
NFRC_FILM_COEFFICIENTS = {
    "exterior": 26.0,  # Winter conditions
    "interior": 7.7,  # Still air
}

# Standard glass conductivity (W/m·K)
GLASS_CONDUCTIVITY = 1.0


@dataclass
class LayerThermalProperties:
    """Thermal properties of a single layer.

    Attributes:
        name: Material name
        obj_type: EnergyPlus object type
        thickness: Layer thickness in meters (None for no-mass layers)
        conductivity: Thermal conductivity in W/m·K (None for no-mass layers)
        r_value: Thermal resistance in m²·K/W
        is_glazing: True if this is a glazing layer
        is_gas: True if this is a gas gap
        gas_type: Gas type if is_gas (Air, Argon, Krypton, Xenon)
        solar_transmittance: Solar transmittance at normal incidence (glazing only)
        solar_reflectance_front: Front-side solar reflectance (glazing only)
        visible_transmittance: Visible transmittance at normal incidence (glazing only)
        emissivity_front: Front-side infrared emissivity (glazing only)
        emissivity_back: Back-side infrared emissivity (glazing only)
        shgc: Solar heat gain coefficient (SimpleGlazingSystem only)
    """

    name: str
    obj_type: str
    thickness: float | None = None
    conductivity: float | None = None
    r_value: float = 0.0
    is_glazing: bool = False
    is_gas: bool = False
    gas_type: str | None = None
    solar_transmittance: float | None = None
    solar_reflectance_front: float | None = None
    visible_transmittance: float | None = None
    emissivity_front: float | None = None
    emissivity_back: float | None = None
    shgc: float | None = None


@dataclass
class ConstructionThermalProperties:
    """Thermal properties of a complete construction assembly.

    Attributes:
        name: Construction name
        layers: List of layer thermal properties (outside to inside)
        r_value: Total thermal resistance in m²·K/W (excluding films)
        r_value_with_films: Total R-value including surface films
        u_value: Overall heat transfer coefficient in W/m²·K (with films)
        is_glazing: True if this is a glazing construction
        shgc: Solar heat gain coefficient (glazing only)
        visible_transmittance: Visible transmittance (glazing only)
    """

    name: str
    layers: list[LayerThermalProperties] = field(default_factory=lambda: [])
    r_value: float = 0.0
    r_value_with_films: float = 0.0
    u_value: float = 0.0
    is_glazing: bool = False
    shgc: float | None = None
    visible_transmittance: float | None = None


def _get_material_by_name(construction: IDFObject, material_name: str) -> IDFObject | None:
    """Get material object by name from the document."""
    document = construction.theidf
    if document is None:
        return None

    for mat_type in _MATERIAL_TYPES:
        if mat_type in document.collections:
            mat = document.collections[mat_type].get(material_name)
            if mat is not None:
                return mat

    return None


def _get_material_layer(
    construction: IDFObject,
    material_name: str,
) -> LayerThermalProperties | None:
    """Get thermal properties for a material layer.

    Args:
        construction: Construction IDFObject (used to access parent document)
        material_name: Name of the material to look up

    Returns:
        LayerThermalProperties or None if material not found
    """
    mat = _get_material_by_name(construction, material_name)
    if mat is None:
        return None
    return _extract_layer_properties(mat)


def _extract_layer_properties(material: IDFObject) -> LayerThermalProperties:
    """Extract thermal properties from a material object.

    Args:
        material: Material IDFObject

    Returns:
        LayerThermalProperties with relevant thermal data
    """
    mat_type = material.obj_type

    if mat_type == "Material":
        thickness = material.thickness if material.thickness is not None else 0.0
        conductivity = material.conductivity if material.conductivity is not None else 1.0
        r_value = thickness / conductivity if conductivity > 0 else 0.0
        return LayerThermalProperties(
            name=material.name,
            obj_type=mat_type,
            thickness=thickness,
            conductivity=conductivity,
            r_value=r_value,
        )

    if mat_type == "Material:NoMass":
        r_value = material.thermal_resistance if material.thermal_resistance is not None else 0.0
        return LayerThermalProperties(
            name=material.name,
            obj_type=mat_type,
            r_value=r_value,
        )

    if mat_type == "Material:AirGap":
        r_value = material.thermal_resistance if material.thermal_resistance is not None else 0.0
        return LayerThermalProperties(
            name=material.name,
            obj_type=mat_type,
            r_value=r_value,
            is_gas=True,
            gas_type="Air",
        )

    if mat_type == "WindowMaterial:Glazing":
        thickness = material.thickness if material.thickness is not None else 0.006  # Default 6mm
        conductivity = material.conductivity if material.conductivity is not None else GLASS_CONDUCTIVITY
        r_value = thickness / conductivity if conductivity > 0 else 0.0

        # Get optical properties
        solar_trans = material.solar_transmittance_at_normal_incidence
        solar_refl_front = material.front_side_solar_reflectance_at_normal_incidence
        vis_trans = material.visible_transmittance_at_normal_incidence
        emiss_front = (
            material.front_side_infrared_hemispherical_emissivity
            if material.front_side_infrared_hemispherical_emissivity is not None
            else 0.84
        )
        emiss_back = (
            material.back_side_infrared_hemispherical_emissivity
            if material.back_side_infrared_hemispherical_emissivity is not None
            else 0.84
        )

        return LayerThermalProperties(
            name=material.name,
            obj_type=mat_type,
            thickness=thickness,
            conductivity=conductivity,
            r_value=r_value,
            is_glazing=True,
            solar_transmittance=solar_trans,
            solar_reflectance_front=solar_refl_front,
            visible_transmittance=vis_trans,
            emissivity_front=emiss_front,
            emissivity_back=emiss_back,
        )

    if mat_type == "WindowMaterial:Gas":
        thickness = material.thickness if material.thickness is not None else 0.012  # Default 12mm
        gas_type = material.gas_type if material.gas_type is not None else "Air"

        # Get R-value from typical values
        r_value = typical_gap_r_value(gas_type, thickness * 1000)

        return LayerThermalProperties(
            name=material.name,
            obj_type=mat_type,
            thickness=thickness,
            r_value=r_value,
            is_gas=True,
            gas_type=gas_type,
        )

    if mat_type == "WindowMaterial:SimpleGlazingSystem":
        # SimpleGlazingSystem provides U-factor directly (includes films)
        u_factor = material.u_factor if material.u_factor is not None else 2.0
        r_value = 1.0 / u_factor if u_factor > 0 else 0.0
        return LayerThermalProperties(
            name=material.name,
            obj_type=mat_type,
            r_value=r_value,
            is_glazing=True,
            shgc=material.solar_heat_gain_coefficient,
            visible_transmittance=material.visible_transmittance,
        )

    # Unknown material type
    return LayerThermalProperties(
        name=material.name,
        obj_type=mat_type,
        r_value=0.0,
    )


def get_construction_layers(construction: IDFObject) -> list[LayerThermalProperties]:
    """Get thermal properties for all layers in a construction.

    Args:
        construction: Construction IDFObject

    Returns:
        List of LayerThermalProperties from outside to inside
    """
    layers: list[LayerThermalProperties] = []

    for field_name in _LAYER_FIELDS:
        material_name = getattr(construction, field_name, None)
        if material_name:
            layer = _get_material_layer(construction, material_name)
            if layer:
                layers.append(layer)

    return layers


def _nfrc_film_resistance() -> float:
    """Total NFRC film resistance (exterior + interior) in m²·K/W."""
    return 1.0 / NFRC_FILM_COEFFICIENTS["exterior"] + 1.0 / NFRC_FILM_COEFFICIENTS["interior"]


def calculate_r_value(construction: IDFObject, include_films: bool = True) -> float:
    """Calculate R-value for a construction assembly.

    For opaque constructions, sums the thermal resistance of all layers.
    For glazing constructions with SimpleGlazingSystem, the stored R-value
    includes films; when ``include_films=False`` the NFRC film resistances
    are subtracted.

    Args:
        construction: Construction IDFObject
        include_films: Whether to include surface film resistances (default True)

    Returns:
        Thermal resistance in m²·K/W
    """
    layers = get_construction_layers(construction)

    if not layers:
        return 0.0

    # Check for SimpleGlazingSystem
    if len(layers) == 1 and layers[0].obj_type == "WindowMaterial:SimpleGlazingSystem":
        # r_value already includes films (1 / U-factor)
        if include_films:
            return layers[0].r_value
        return max(0.0, layers[0].r_value - _nfrc_film_resistance())

    # Sum layer resistances
    r_total = sum(layer.r_value for layer in layers)

    # Add film resistances
    if include_films:
        # Determine if this is a glazing construction
        is_glazing = any(layer.is_glazing for layer in layers)

        if is_glazing:
            # Use NFRC film coefficients for windows
            r_exterior = 1.0 / NFRC_FILM_COEFFICIENTS["exterior"]
            r_interior = 1.0 / NFRC_FILM_COEFFICIENTS["interior"]
        else:
            # Use ASHRAE film resistances for opaque
            r_exterior = FILM_RESISTANCE["exterior"]
            r_interior = FILM_RESISTANCE["interior"]

        r_total += r_exterior + r_interior

    return r_total


def calculate_u_value(construction: IDFObject) -> float:
    """Calculate U-value for a construction assembly.

    Args:
        construction: Construction IDFObject

    Returns:
        Overall heat transfer coefficient in W/m²·K
    """
    r_value = calculate_r_value(construction, include_films=True)
    if r_value > 0:
        return 1.0 / r_value
    return 0.0


def calculate_shgc(construction: IDFObject) -> float | None:
    """Calculate Solar Heat Gain Coefficient for a glazing construction.

    Uses simplified calculation based on solar transmittance and absorptance.
    For accurate values, use the full WINDOW/TARCOG algorithm.

    Args:
        construction: Construction IDFObject (must be glazing)

    Returns:
        SHGC value (0-1) or None if not a glazing construction
    """
    layers = get_construction_layers(construction)

    if not layers:
        return None

    # Check for SimpleGlazingSystem - SHGC is provided directly
    if len(layers) == 1 and layers[0].obj_type == "WindowMaterial:SimpleGlazingSystem":
        return layers[0].shgc

    # Check if this is a glazing construction
    glazing_layers = [layer for layer in layers if layer.is_glazing]
    if not glazing_layers:
        return None

    # Simplified SHGC calculation for multi-pane systems
    # Product of individual transmittances for direct component
    tau_direct = 1.0
    for layer in glazing_layers:
        if layer.solar_transmittance is not None:
            tau_direct *= layer.solar_transmittance

    # Absorbed component contribution (simplified)
    # For the outermost glazing layer, some absorbed radiation flows inward
    first_glazing = glazing_layers[0]
    if first_glazing.solar_transmittance is not None and first_glazing.solar_reflectance_front is not None:
        tau = first_glazing.solar_transmittance
        rho = first_glazing.solar_reflectance_front
        alpha = 1.0 - tau - rho

        # Inward-flowing fraction (simplified)
        # Uses NFRC film coefficients
        h_out = NFRC_FILM_COEFFICIENTS["exterior"]
        h_in = NFRC_FILM_COEFFICIENTS["interior"]
        n_i = alpha * (h_in / (h_out + h_in))

        shgc = tau_direct + n_i
        return min(shgc, 1.0)

    return tau_direct


def calculate_visible_transmittance(construction: IDFObject) -> float | None:
    """Calculate visible transmittance for a glazing construction.

    Args:
        construction: Construction IDFObject (must be glazing)

    Returns:
        Visible transmittance (0-1) or None if not a glazing construction
    """
    layers = get_construction_layers(construction)

    if not layers:
        return None

    # Check for SimpleGlazingSystem - VT is provided directly
    if len(layers) == 1 and layers[0].obj_type == "WindowMaterial:SimpleGlazingSystem":
        return layers[0].visible_transmittance

    # Product of individual visible transmittances
    vt = 1.0
    has_glazing = False

    for layer in layers:
        if layer.is_glazing and layer.visible_transmittance is not None:
            has_glazing = True
            vt *= layer.visible_transmittance

    return vt if has_glazing else None


def get_thermal_properties(construction: IDFObject) -> ConstructionThermalProperties:
    """Get complete thermal properties for a construction.

    Args:
        construction: Construction IDFObject

    Returns:
        ConstructionThermalProperties with all calculated values
    """
    layers = get_construction_layers(construction)

    # Determine if glazing
    is_glazing = any(layer.is_glazing for layer in layers)

    # Calculate R-values
    r_value = calculate_r_value(construction, include_films=False)
    r_value_with_films = calculate_r_value(construction, include_films=True)

    # Calculate U-value
    u_value = 1.0 / r_value_with_films if r_value_with_films > 0 else 0.0

    # Calculate SHGC and VT for glazing
    shgc = calculate_shgc(construction) if is_glazing else None
    vt = calculate_visible_transmittance(construction) if is_glazing else None

    return ConstructionThermalProperties(
        name=construction.name,
        layers=layers,
        r_value=r_value,
        r_value_with_films=r_value_with_films,
        u_value=u_value,
        is_glazing=is_glazing,
        shgc=shgc,
        visible_transmittance=vt,
    )



from __future__ import annotations

from idfkit.simulation import SQLResult, SimulationResult, TimeSeriesResult

result: SimulationResult = ...  # type: ignore[assignment]
sql: SQLResult | None = ...  # type: ignore[assignment]
ts: TimeSeriesResult = ...  # type: ignore[assignment]
# --8<-- [start:example]
sql = result.sql
if sql is not None:
    # Time-series data
    ts = sql.get_timeseries(
        variable_name="Zone Mean Air Temperature",
        key_value="THERMAL ZONE 1",
    )
    print(f"Max: {max(ts.values):.1f}°C")

    # Tabular reports
    rows = sql.get_tabular_data(report_name="AnnualBuildingUtilityPerformanceSummary")
# --8<-- [end:example]



from __future__ import annotations

# --8<-- [start:example]
import logging

# Show all INFO-level messages from idfkit
logging.basicConfig(level=logging.INFO)

# Load and parse — idfkit logs progress automatically
from idfkit import load_idf

doc = load_idf("model.idf")
# INFO:idfkit.idf_parser:Parsed 850 objects from model.idf in 0.142s
# --8<-- [end:example]



"""Tests for idfkit.weather.geocode (mocked, no network)."""

from __future__ import annotations

import json
from unittest.mock import MagicMock, patch

import pytest

from idfkit.weather.geocode import GeocodingError, RateLimiter, _nominatim_limiter, geocode


def _mock_response(data: bytes) -> MagicMock:
    mock_resp = MagicMock()
    mock_resp.read.return_value = data
    mock_resp.__enter__ = MagicMock(return_value=mock_resp)
    mock_resp.__exit__ = MagicMock(return_value=False)
    return mock_resp


@pytest.fixture(autouse=True)
def reset_rate_limiter() -> None:
    """Reset the global rate limiter before each test."""
    _nominatim_limiter.reset()


class TestRateLimiter:
    """Tests for the RateLimiter class."""

    def test_first_call_no_wait(self) -> None:
        limiter = RateLimiter(min_interval=1.0)
        with patch("idfkit.weather.geocode.time") as mock_time:
            mock_time.monotonic.return_value = 100.0
            mock_time.sleep = MagicMock()
            limiter.wait()
            mock_time.sleep.assert_not_called()

    def test_subsequent_call_waits(self) -> None:
        limiter = RateLimiter(min_interval=1.0)
        with patch("idfkit.weather.geocode.time") as mock_time:
            # First call at t=100.0
            mock_time.monotonic.return_value = 100.0
            mock_time.sleep = MagicMock()
            limiter.wait()

            # Second call at t=100.3 (only 0.3s elapsed)
            mock_time.monotonic.return_value = 100.3
            limiter.wait()
            mock_time.sleep.assert_called_once()
            # Should sleep for 0.7s to reach the 1.0s interval
            sleep_duration = mock_time.sleep.call_args[0][0]
            assert abs(sleep_duration - 0.7) < 0.01

    def test_no_wait_after_interval(self) -> None:
        limiter = RateLimiter(min_interval=1.0)
        with patch("idfkit.weather.geocode.time") as mock_time:
            # First call at t=100.0
            mock_time.monotonic.return_value = 100.0
            mock_time.sleep = MagicMock()
            limiter.wait()

            # Second call at t=101.5 (1.5s elapsed, > 1.0s interval)
            mock_time.monotonic.return_value = 101.5
            limiter.wait()
            mock_time.sleep.assert_not_called()

    def test_reset(self) -> None:
        limiter = RateLimiter(min_interval=1.0)
        with patch("idfkit.weather.geocode.time") as mock_time:
            mock_time.monotonic.return_value = 100.0
            mock_time.sleep = MagicMock()
            limiter.wait()

            # After reset, no waiting should occur
            limiter.reset()
            mock_time.monotonic.return_value = 100.1
            limiter.wait()
            mock_time.sleep.assert_not_called()


class TestGeocode:
    @patch("urllib.request.urlopen")
    def test_successful_geocode(self, mock_urlopen: MagicMock) -> None:
        response_data = [{"lat": "41.8781", "lon": "-87.6298", "display_name": "Chicago"}]
        mock_urlopen.return_value = _mock_response(json.dumps(response_data).encode())

        lat, lon = geocode("Willis Tower, Chicago, IL")
        assert abs(lat - 41.8781) < 0.001
        assert abs(lon - (-87.6298)) < 0.001

    @patch("urllib.request.urlopen")
    def test_empty_response_raises(self, mock_urlopen: MagicMock) -> None:
        mock_urlopen.return_value = _mock_response(b"[]")

        with pytest.raises(GeocodingError, match="No results found"):
            geocode("zzzznonexistentplace")

    @patch("urllib.request.urlopen")
    def test_network_error_raises(self, mock_urlopen: MagicMock) -> None:
        from urllib.error import URLError

        mock_urlopen.side_effect = URLError("Network unreachable")

        with pytest.raises(GeocodingError, match="Failed to geocode"):
            geocode("Chicago, IL")

    @patch("urllib.request.urlopen")
    def test_invalid_json_raises(self, mock_urlopen: MagicMock) -> None:
        mock_urlopen.return_value = _mock_response(b"not valid json")

        with pytest.raises(GeocodingError, match="Failed to geocode"):
            geocode("Some address")

    @patch("urllib.request.urlopen")
    def test_missing_lat_lon_raises(self, mock_urlopen: MagicMock) -> None:
        response_data = [{"display_name": "Chicago"}]  # Missing lat/lon
        mock_urlopen.return_value = _mock_response(json.dumps(response_data).encode())

        with pytest.raises(GeocodingError, match="Failed to geocode"):
            geocode("Chicago, IL")



from __future__ import annotations

from idfkit import IDFDocument
from idfkit.simulation import OutputVariableIndex

model: IDFDocument = ...  # type: ignore[assignment]
variables: OutputVariableIndex | None = ...  # type: ignore[assignment]
# --8<-- [start:example]
# Search first, review, then add selectively
matches = variables.search("Heating")

# Filter to specific ones
selected = [v for v in matches if "Coil" in v.name]

# Add to model (name is optional for Output:Variable)
for var in selected:
    model.add(
        "Output:Variable",
        key_value="*",
        variable_name=var.name,
        reporting_frequency="Timestep",
    )
# --8<-- [end:example]



"""Batch simulation execution with thread-pool parallelism.

Runs multiple EnergyPlus simulations concurrently using a
[ThreadPoolExecutor][concurrent.futures.ThreadPoolExecutor].  Thread-based parallelism
is appropriate because [simulate][idfkit.simulation.runner.simulate] delegates to ``subprocess.run``
which releases the GIL.
"""

from __future__ import annotations

import logging
import os
import tempfile
import time
from collections.abc import Callable, Sequence
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Literal

from .progress import SimulationProgress
from .progress_bars import resolve_on_progress
from .result import SimulationResult
from .runner import simulate

if TYPE_CHECKING:
    from .cache import SimulationCache
    from .config import EnergyPlusConfig
    from .fs import FileSystem

logger = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class SimulationJob:
    """Specification for a single simulation within a batch.

    Attributes:
        model: The EnergyPlus model to simulate.
        weather: Path to the weather file.
        label: Human-readable label for progress reporting.
        output_dir: Directory for output files (default: auto temp dir).
        expand_objects: Run ExpandObjects before simulation.
        annual: Run annual simulation.
        design_day: Run design-day-only simulation.
        output_prefix: Prefix for output files.
        output_suffix: Output file naming suffix (``"C"``, ``"L"``, or ``"D"``).
        readvars: Run ReadVarsESO after simulation.
        timeout: Maximum runtime in seconds.
        extra_args: Additional command-line arguments.
    """

    model: object  # IDFDocument — use object to avoid import at class level
    weather: str | Path
    label: str = ""
    output_dir: str | Path | None = None
    expand_objects: bool = True
    annual: bool = False
    design_day: bool = False
    output_prefix: str = "eplus"
    output_suffix: Literal["C", "L", "D"] = "C"
    readvars: bool = False
    timeout: float = 3600.0
    extra_args: tuple[str, ...] | None = None


@dataclass(frozen=True, slots=True)
class BatchResult:
    """Aggregated results from a batch simulation run.

    Attributes:
        results: Simulation results in the same order as the input jobs.
        total_runtime_seconds: Wall-clock time for the entire batch.
    """

    results: tuple[SimulationResult, ...]
    total_runtime_seconds: float

    @property
    def succeeded(self) -> tuple[SimulationResult, ...]:
        """Results that completed successfully."""
        return tuple(r for r in self.results if r.success)

    @property
    def failed(self) -> tuple[SimulationResult, ...]:
        """Results that failed."""
        return tuple(r for r in self.results if not r.success)

    @property
    def all_succeeded(self) -> bool:
        """Whether every job in the batch succeeded."""
        return all(r.success for r in self.results)

    def __len__(self) -> int:
        return len(self.results)

    def __getitem__(self, index: int) -> SimulationResult:
        return self.results[index]


def simulate_batch(
    jobs: Sequence[SimulationJob],
    *,
    energyplus: EnergyPlusConfig | None = None,
    max_workers: int | None = None,
    cache: SimulationCache | None = None,
    progress: Callable[..., None] | None = None,
    fs: FileSystem | None = None,
    on_progress: Callable[[SimulationProgress], None] | None = None,
) -> BatchResult:
    """Run multiple EnergyPlus simulations in parallel.

    Uses [ThreadPoolExecutor][concurrent.futures.ThreadPoolExecutor] to dispatch
    simulations concurrently.  Individual job failures are captured as
    failed [SimulationResult][idfkit.simulation.result.SimulationResult] entries -- the batch never raises
    due to a single job failing.

    Args:
        jobs: Sequence of simulation jobs to execute.
        energyplus: Shared EnergyPlus configuration (auto-discovered if
            ``None``).
        max_workers: Maximum number of concurrent simulations.  Defaults
            to ``min(len(jobs), os.cpu_count() or 1)``.
        cache: Optional simulation cache for content-hash lookups.
        progress: Optional callback invoked after each job completes.
            Called as ``progress(completed=N, total=M, label=label,
            success=bool)``.
        fs: Optional file system backend passed through to each
            [simulate][idfkit.simulation.runner.simulate] call.
        on_progress: Optional callback invoked with
            [SimulationProgress][idfkit.simulation.progress.SimulationProgress] events
            during each individual simulation.  Events include
            ``job_index`` and ``job_label`` to identify which batch job
            they belong to.  The ``"tqdm"`` shorthand is not supported
            for batch runners; use
            [tqdm_progress][idfkit.simulation.progress_bars.tqdm_progress]
            with a custom per-job callback instead.

    Returns:
        A [BatchResult][idfkit.simulation.batch.BatchResult] with results in the same order as *jobs*.

    Raises:
        ValueError: If *jobs* is empty.
    """
    if not jobs:
        msg = "jobs must not be empty"
        raise ValueError(msg)

    if on_progress == "tqdm":
        msg = (
            'on_progress="tqdm" is not supported for batch simulations because a single '
            "progress bar cannot represent multiple concurrent jobs. Use the tqdm_progress() "
            "context manager with a custom callback instead."
        )
        raise ValueError(msg)

    progress_cb, progress_cleanup = resolve_on_progress(on_progress)

    try:
        if max_workers is None:
            max_workers = min(len(jobs), os.cpu_count() or 1)

        logger.info("Starting batch of %d jobs with %d workers", len(jobs), max_workers)

        results: list[SimulationResult | None] = [None] * len(jobs)
        completed_count = 0
        total = len(jobs)

        start = time.monotonic()

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            future_to_index = {
                executor.submit(_run_job, idx, job, energyplus, cache, fs, progress_cb): idx
                for idx, job in enumerate(jobs)
            }

            for future in as_completed(future_to_index):
                idx = future_to_index[future]
                job = jobs[idx]
                result = future.result()  # _run_job never raises
                results[idx] = result
                completed_count += 1

                if progress is not None:
                    progress(
                        completed=completed_count,
                        total=total,
                        label=job.label,
                        success=result.success,
                    )
    finally:
        if progress_cleanup is not None:
            progress_cleanup()

    elapsed = time.monotonic() - start

    # All slots filled — assert for type checker
    final: list[SimulationResult] = []
    for r in results:
        assert r is not None  # noqa: S101
        final.append(r)

    batch_result = BatchResult(results=tuple(final), total_runtime_seconds=elapsed)
    logger.info(
        "Batch complete: %d succeeded, %d failed in %.1fs",
        len(batch_result.succeeded),
        len(batch_result.failed),
        elapsed,
    )
    return batch_result


def _run_job(
    idx: int,
    job: SimulationJob,
    energyplus: EnergyPlusConfig | None,
    cache: SimulationCache | None,
    fs: FileSystem | None,
    on_progress: Callable[[SimulationProgress], None] | None,
) -> SimulationResult:
    """Execute a single simulation job, catching SimulationError."""
    # Wrap the user callback to inject batch job context
    job_cb: Callable[[SimulationProgress], None] | None = None
    if on_progress is not None:
        job_cb = _make_job_progress_callback(on_progress, idx, job.label)

    try:
        return simulate(
            model=job.model,  # type: ignore[arg-type]
            weather=job.weather,
            output_dir=job.output_dir,
            energyplus=energyplus,
            expand_objects=job.expand_objects,
            annual=job.annual,
            design_day=job.design_day,
            output_prefix=job.output_prefix,
            output_suffix=job.output_suffix,
            readvars=job.readvars,
            timeout=job.timeout,
            extra_args=list(job.extra_args) if job.extra_args else None,
            cache=cache,
            fs=fs,
            on_progress=job_cb,
        )
    except Exception as exc:
        # Catch all exceptions (SimulationError, ExpandObjectsError,
        # EnergyPlusNotFoundError, etc.) so that a single job failure
        # never crashes the entire batch.
        failed_run_dir = Path(job.output_dir) if job.output_dir is not None else Path(tempfile.mkdtemp())
        exit_code = getattr(exc, "exit_code", None)
        stderr = getattr(exc, "stderr", None) or str(exc)
        return SimulationResult(
            run_dir=failed_run_dir,
            success=False,
            exit_code=exit_code,
            stdout="",
            stderr=stderr,
            runtime_seconds=0.0,
            output_prefix=job.output_prefix,
        )


def _make_job_progress_callback(
    user_cb: Callable[[SimulationProgress], None],
    job_index: int,
    job_label: str,
) -> Callable[[SimulationProgress], None]:
    """Create a wrapper that stamps each event with batch job context."""

    def _wrapper(event: SimulationProgress) -> None:
        stamped = SimulationProgress(
            phase=event.phase,
            message=event.message,
            percent=event.percent,
            environment=event.environment,
            warmup_day=event.warmup_day,
            sim_day=event.sim_day,
            sim_total_days=event.sim_total_days,
            job_index=job_index,
            job_label=job_label,
        )
        user_cb(stamped)

    return _wrapper



from __future__ import annotations

from idfkit.weather import StationIndex

index: StationIndex = ...  # type: ignore[assignment]
# --8<-- [start:example]
# City name
results = index.search("New York")

# Airport code pattern
results = index.search("JFK")

# WMO number
results = index.search("725300")

# Country + city
results = index.search("London UK")
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument, IDFObject

idf: IDFDocument = ...  # type: ignore[assignment]
surface: IDFObject = ...  # type: ignore[assignment]
# --8<-- [start:example]
surface = idf.idfobjects["BuildingSurface:Detailed"][0]
construction = surface.get_referenced_object("Construction_Name")
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import SimulationResult
from idfkit.weather import StationIndex

index: StationIndex = ...  # type: ignore[assignment]
results: list[SimulationResult | None] = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.weather import StationIndex, geocode

index = StationIndex.load()
results = index.nearest(*geocode("Chicago, IL"))
print(results[0].station.display_name)
# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument
from idfkit.simulation import SimulationResult

model: IDFDocument = ...  # type: ignore[assignment]
result: SimulationResult = ...  # type: ignore[assignment]
# --8<-- [start:example]
from idfkit.simulation import simulate

result = simulate(model, "weather.epw", design_day=True)
print(f"Outputs in: {result.run_dir}")
# --8<-- [end:example]



from __future__ import annotations

from datetime import datetime
from idfkit import IDFDocument, IDFObject
from typing import Any

day_schedule_type: Any = ...  # type: ignore[assignment]
doc: IDFDocument = ...  # type: ignore[assignment]
dt: Any = ...  # type: ignore[assignment]
evaluate_day: Any = ...  # type: ignore[assignment]
evaluate_week: Any = ...  # type: ignore[assignment]
field_name_for_index: Any = ...  # type: ignore[assignment]
find_week_for_date: Any = ...  # type: ignore[assignment]
obj: IDFObject = ...  # type: ignore[assignment]
week_type: Any = ...  # type: ignore[assignment]


# --8<-- [start:example]
def evaluate_year(obj: IDFObject, dt: datetime, doc: IDFDocument) -> float:
    # 1. Find which date range contains dt
    # 2. Get the referenced week schedule name
    # 3. Look up week schedule in document
    # 4. Evaluate week schedule for dt
    week_name = find_week_for_date(obj, dt)
    week_obj = doc.get_schedule(week_name) or doc[week_type][week_name]
    return evaluate_week(week_obj, dt, doc)


def evaluate_week_daily(obj: IDFObject, dt: datetime, doc: IDFDocument) -> float:
    # Schedule:Week:Daily has 12 fields: Sunday-Saturday + Holiday + Summer/Winter DD + Custom
    day_index = dt.weekday()  # 0=Mon, need to map to E+ order (Sun=0)
    field_map = {6: 0, 0: 1, 1: 2, 2: 3, 3: 4, 4: 5, 5: 6}  # Python weekday → E+ field
    day_name = obj[field_name_for_index(field_map[day_index])]
    day_obj = doc[day_schedule_type][day_name]
    return evaluate_day(day_obj, dt, doc)


# --8<-- [end:example]



from __future__ import annotations

from idfkit import IDFDocument
from idfkit.weather import WeatherDownloader, WeatherFiles, WeatherStation

downloader: WeatherDownloader = ...  # type: ignore[assignment]
files: WeatherFiles = ...  # type: ignore[assignment]
model: IDFDocument = ...  # type: ignore[assignment]
station: WeatherStation = ...  # type: ignore[assignment]
# --8<-- [start:example]
files = downloader.download(station)

# Use for simulation
from idfkit.simulation import simulate

result = simulate(model, files.epw)

# Use for design days
from idfkit.weather import DesignDayManager

ddm = DesignDayManager(files.ddy)
# --8<-- [end:example]



from __future__ import annotations

from idfkit.simulation import SQLResult

sql: SQLResult | None = ...  # type: ignore[assignment]
# --8<-- [start:example]
# Design day results only (use for design_day=True simulations)
ts = sql.get_timeseries(
    "Zone Mean Air Temperature",
    "ZONE 1",
    environment="sizing",
)

# Annual/run period results only (default)
ts = sql.get_timeseries(
    "Zone Mean Air Temperature",
    "ZONE 1",
    environment="annual",
)

# All environments (design days + run periods)
ts = sql.get_timeseries(
    "Zone Mean Air Temperature",
    "ZONE 1",
    environment=None,
)
# --8<-- [end:example]



"""
EnergyPlus version registry.

Defines all supported EnergyPlus versions since v8.9 (the first to publish
epJSON schema) and provides utilities for version manipulation.
"""

from __future__ import annotations

from typing import Final

# All EnergyPlus versions since v8.9.0 that publish an epJSON schema.
# Each entry is a (major, minor, patch) tuple.
ENERGYPLUS_VERSIONS: Final[tuple[tuple[int, int, int], ...]] = (
    (8, 9, 0),
    (9, 0, 1),
    (9, 1, 0),
    (9, 2, 0),
    (9, 3, 0),
    (9, 4, 0),
    (9, 5, 0),
    (9, 6, 0),
    (22, 1, 0),
    (22, 2, 0),
    (23, 1, 0),
    (23, 2, 0),
    (24, 1, 0),
    (24, 2, 0),
    (25, 1, 0),
    (25, 2, 0),
)

#: The latest supported EnergyPlus version.
LATEST_VERSION: Final[tuple[int, int, int]] = ENERGYPLUS_VERSIONS[-1]

#: Minimum supported version (first with epJSON schema).
MINIMUM_VERSION: Final[tuple[int, int, int]] = ENERGYPLUS_VERSIONS[0]

# Set for O(1) membership checks
_VERSION_SET: Final[frozenset[tuple[int, int, int]]] = frozenset(ENERGYPLUS_VERSIONS)

# Mapping from GitHub release tags to version tuples
_TAG_TO_VERSION: Final[dict[str, tuple[int, int, int]]] = {f"v{v[0]}.{v[1]}.{v[2]}": v for v in ENERGYPLUS_VERSIONS}

# Mapping from version tuple to the directory name used for bundled schemas
_VERSION_TO_DIRNAME: Final[dict[tuple[int, int, int], str]] = {v: f"V{v[0]}-{v[1]}-{v[2]}" for v in ENERGYPLUS_VERSIONS}


def is_supported_version(version: tuple[int, int, int]) -> bool:
    """Check if a version is in the supported set.

    Examples:
        >>> is_supported_version((24, 1, 0))
        True
        >>> is_supported_version((99, 0, 0))
        False
        >>> is_supported_version(MINIMUM_VERSION)
        True
    """
    return version in _VERSION_SET


def version_string(version: tuple[int, int, int]) -> str:
    """Format a version tuple as a human-readable string.

    Examples:
        >>> version_string((24, 1, 0))
        '24.1.0'
        >>> version_string((9, 6, 0))
        '9.6.0'
    """
    return f"{version[0]}.{version[1]}.{version[2]}"


def version_dirname(version: tuple[int, int, int]) -> str:
    """Return the schema directory name for a version.

    Examples:
        >>> version_dirname((24, 1, 0))
        'V24-1-0'
        >>> version_dirname((9, 6, 0))
        'V9-6-0'
    """
    return f"V{version[0]}-{version[1]}-{version[2]}"


def find_closest_version(version: tuple[int, int, int]) -> tuple[int, int, int] | None:
    """
    Find the closest supported version that is <= the given version.

    This is useful when a file specifies a patch version that doesn't
    exactly match a supported version (e.g. 9.0.0 -> 9.0.1).

    Returns:
        The closest supported version, or None if no suitable version exists.

    Examples:
        >>> find_closest_version((24, 1, 5))
        (24, 1, 0)
        >>> find_closest_version((9, 0, 0))
        (8, 9, 0)
        >>> find_closest_version((1, 0, 0)) is None
        True
    """
    best: tuple[int, int, int] | None = None
    for v in ENERGYPLUS_VERSIONS:
        if v <= version:
            best = v
    return best


def github_release_tag(version: tuple[int, int, int]) -> str:
    """Return the GitHub release tag for a version.

    Examples:
        >>> github_release_tag((24, 1, 0))
        'v24.1.0'
        >>> github_release_tag((9, 2, 0))
        'v9.2.0'
    """
    return f"v{version[0]}.{version[1]}.{version[2]}"
